"""
:mod: DataManager

.. module: DataManager

:synopsis: DataManager links the functionalities of StorageElement and FileCatalog.

This module consists of DataManager and related classes.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# # imports
from datetime import datetime, timedelta
import fnmatch
import os
import time
import errno
import six

# # from DIRAC
import DIRAC
from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities.Adler import fileAdler, compareAdler
from DIRAC.Core.Utilities.File import makeGuid, getSize
from DIRAC.Core.Utilities.List import randomize, breakListIntoChunks
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.AccountingSystem.Client.DataStoreClient import gDataStoreClient
from DIRAC.AccountingSystem.Client.Types.DataOperation import DataOperation
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus

# # RSCID
__RCSID__ = "$Id$"


def _isOlderThan(stringTime, days):
    """Check if a time stamp is older than a given number of days"""
    timeDelta = timedelta(days=days)
    maxCTime = datetime.utcnow() - timeDelta
    # st = time.strptime( stringTime, "%a %b %d %H:%M:%S %Y" )
    # cTimeStruct = datetime( st[0], st[1], st[2], st[3], st[4], st[5], st[6], None )
    cTimeStruct = stringTime
    if cTimeStruct < maxCTime:
        return True
    return False


def _initialiseAccountingObject(operation, se, files):
    """create accouting record"""
    accountingDict = {}
    accountingDict["OperationType"] = operation
    result = getProxyInfo()
    if not result["OK"]:
        userName = "system"
    else:
        userName = result["Value"].get("username", "unknown")
    accountingDict["User"] = userName
    accountingDict["Protocol"] = "DataManager"
    accountingDict["RegistrationTime"] = 0.0
    accountingDict["RegistrationOK"] = 0
    accountingDict["RegistrationTotal"] = 0
    accountingDict["Destination"] = se
    accountingDict["TransferTotal"] = files
    accountingDict["TransferOK"] = files
    accountingDict["TransferSize"] = files
    accountingDict["TransferTime"] = 0.0
    accountingDict["FinalStatus"] = "Successful"
    accountingDict["Source"] = DIRAC.siteName()
    oDataOperation = DataOperation()
    oDataOperation.setValuesFromDict(accountingDict)
    return oDataOperation


class DataManager(object):
    """
    .. class:: DataManager

    A DataManager is taking all the actions that impact or require the FileCatalog and the StorageElement together
    """

    def __init__(self, catalogs=None, masterCatalogOnly=False, vo=False):
        """c'tor

        :param self: self reference
        :param catalogs: the list of catalog in which to perform the operations. This
                        list will be ignored if masterCatalogOnly is set to True
        :param masterCatalogOnly: if set to True, the operations will be performed only on the master catalog.
                                  The catalogs parameter will be ignored.
        :param vo: the VO for which the DataManager is created, get VO from the current proxy if not specified
        """
        self.log = gLogger.getSubLogger(self.__class__.__name__)
        self.voName = vo

        if catalogs is None:
            catalogs = []
        catalogsToUse = FileCatalog(vo=self.voName).getMasterCatalogNames()["Value"] if masterCatalogOnly else catalogs

        self.fileCatalog = FileCatalog(catalogs=catalogsToUse, vo=self.voName)
        self.accountingClient = None
        self.resourceStatus = ResourceStatus()
        self.ignoreMissingInFC = Operations(vo=self.voName).getValue("DataManagement/IgnoreMissingInFC", False)
        self.useCatalogPFN = Operations(vo=self.voName).getValue("DataManagement/UseCatalogPFN", True)
        self.dmsHelper = DMSHelpers(vo=vo)
        self.registrationProtocol = self.dmsHelper.getRegistrationProtocols()
        self.thirdPartyProtocols = self.dmsHelper.getThirdPartyProtocols()

    def setAccountingClient(self, client):
        """Set Accounting Client instance"""
        self.accountingClient = client

    def __hasAccess(self, opType, path):
        """Check if we have permission to execute given operation on the given file (if exists) or its directory"""
        if isinstance(path, six.string_types):
            paths = [path]
        else:
            paths = list(path)

        res = self.fileCatalog.hasAccess(paths, opType)
        if not res["OK"]:
            return res
        result = {"Successful": list(), "Failed": list()}
        for path in paths:
            isAllowed = res["Value"]["Successful"].get(path, False)
            if isAllowed:
                result["Successful"].append(path)
            else:
                result["Failed"].append(path)
        return S_OK(result)

    ##########################################################################
    #
    # These are the bulk removal methods
    #

    def cleanLogicalDirectory(self, lfnDir):
        """Clean the logical directory from the catalog and storage"""
        log = self.log.getSubLogger("cleanLogicalDirectory")
        if isinstance(lfnDir, six.string_types):
            lfnDir = [lfnDir]
        retDict = {"Successful": {}, "Failed": {}}
        for folder in lfnDir:
            res = self.__cleanDirectory(folder)
            if not res["OK"]:
                log.debug("Failed to clean directory.", "%s %s" % (folder, res["Message"]))
                retDict["Failed"][folder] = res["Message"]
            else:
                log.debug("Successfully removed directory.", folder)
                retDict["Successful"][folder] = res["Value"]
        return S_OK(retDict)

    def __cleanDirectory(self, folder):
        """delete all files from directory :folder: in FileCatalog and StorageElement

        :param self: self reference
        :param str folder: directory name
        """
        log = self.log.getSubLogger("__cleanDirectory")
        res = self.__hasAccess("removeDirectory", folder)
        if not res["OK"]:
            return res
        if folder not in res["Value"]["Successful"]:
            errStr = "Write access not permitted for this credential."
            log.debug(errStr, folder)
            return S_ERROR(errStr)
        res = self.__getCatalogDirectoryContents([folder], includeDirectories=True)
        if not res["OK"]:
            return res

        if not res["Value"]:
            # folder is empty, just remove it and return
            res = returnSingleResult(self.fileCatalog.removeDirectory(folder, recursive=True))
            return res

        # create a list of folders so that empty folders are also deleted
        areDirs = self.fileCatalog.isDirectory(res["Value"])
        if not areDirs["OK"]:
            return areDirs
        listOfFolders = [aDir for aDir in areDirs["Value"]["Successful"] if areDirs["Value"]["Successful"][aDir]]
        for lfn in listOfFolders:
            res["Value"].pop(lfn)

        res = self.removeFile(res["Value"])
        if not res["OK"]:
            return res
        for lfn, reason in res["Value"]["Failed"].items():  # can be an iterator
            log.error("Failed to remove file found in the catalog", "%s %s" % (lfn, reason))
        res = returnSingleResult(self.removeFile(["%s/dirac_directory" % folder]))
        if not res["OK"]:
            if not DErrno.cmpError(res, errno.ENOENT):
                log.warn("Failed to delete dirac_directory placeholder file")

        storageElements = gConfig.getValue("Resources/StorageElementGroups/SE_Cleaning_List", [])
        failed = False
        for storageElement in sorted(storageElements):
            res = self.__removeStorageDirectory(folder, storageElement)
            if not res["OK"]:
                failed = True
        if failed:
            return S_ERROR("Failed to clean storage directory at all SEs")

        for aFolder in sorted(listOfFolders, reverse=True):
            res = returnSingleResult(self.fileCatalog.removeDirectory(aFolder, recursive=True))
            log.verbose("Removed folder", "%s: %s" % (aFolder, res))
            if not res["OK"]:
                return res

        res = returnSingleResult(self.fileCatalog.removeDirectory(folder, recursive=True))
        if not res["OK"]:
            return res
        return S_OK()

    def __removeStorageDirectory(self, directory, storageElement):
        """delete SE directory

        :param self: self reference
        :param str directory: folder to be removed
        :param str storageElement: DIRAC SE name
        """

        se = StorageElement(storageElement, vo=self.voName)
        res = returnSingleResult(se.exists(directory))

        log = self.log.getSubLogger("__removeStorageDirectory")
        if not res["OK"]:
            log.debug("Failed to obtain existance of directory", res["Message"])
            return res

        exists = res["Value"]
        if not exists:
            log.debug("The directory %s does not exist at %s " % (directory, storageElement))
            return S_OK()

        res = returnSingleResult(se.removeDirectory(directory, recursive=True))
        if not res["OK"]:
            log.debug("Failed to remove storage directory", res["Message"])
            return res

        log.debug(
            "Successfully removed %d files from %s at %s" % (res["Value"]["FilesRemoved"], directory, storageElement)
        )
        return S_OK()

    def __getCatalogDirectoryContents(self, directories, includeDirectories=False):
        """ls recursively all files in directories

        :param self: self reference
        :param list directories: folder names
        :param bool includeDirectories: if True includes directories in the return dictionary
        :return: S_OK with dict of LFNs and their attribute dictionary
        """
        log = self.log.getSubLogger("__getCatalogDirectoryContents")
        log.debug("Obtaining the catalog contents for %d directories:" % len(directories))
        activeDirs = directories
        allFiles = {}
        while len(activeDirs) > 0:
            currentDir = activeDirs[0]
            res = returnSingleResult(self.fileCatalog.listDirectory(currentDir, verbose=True))
            activeDirs.remove(currentDir)

            if not res["OK"]:
                log.debug("Problem getting the %s directory content" % currentDir, res["Message"])
            else:
                dirContents = res["Value"]
                activeDirs.extend(dirContents["SubDirs"])
                allFiles.update(dirContents["Files"])
                if includeDirectories:
                    allFiles.update(dirContents["SubDirs"])
        log.debug("Found %d files" % len(allFiles))
        return S_OK(allFiles)

    def getReplicasFromDirectory(self, directory):
        """get all replicas from a given directory

        :param self: self reference
        :param mixed directory: list of directories or one directory
        """
        if isinstance(directory, six.string_types):
            directories = [directory]
        else:
            directories = directory
        res = self.__getCatalogDirectoryContents(directories)
        if not res["OK"]:
            return res
        allReplicas = dict((lfn, metadata["Replicas"]) for lfn, metadata in res["Value"].items())  # can be an iterator
        return S_OK(allReplicas)

    def getFilesFromDirectory(self, directory, days=0, wildcard="*"):
        """get all files from :directory: older than :days: days matching to :wildcard:

        :param self: self reference
        :param mixed directory: list of directories or directory name
        :param int days: ctime days
        :param str wildcard: pattern to match
        """
        if isinstance(directory, six.string_types):
            directories = [directory]
        else:
            directories = directory
        log = self.log.getSubLogger("getFilesFromDirectory")
        log.debug("Obtaining the files older than %d days in %d directories:" % (days, len(directories)))
        for folder in directories:
            log.debug(folder)
        activeDirs = directories
        allFiles = []
        while len(activeDirs) > 0:
            currentDir = activeDirs[0]
            # We only need the metadata (verbose) if a limit date is given
            res = returnSingleResult(self.fileCatalog.listDirectory(currentDir, verbose=(days != 0)))
            activeDirs.remove(currentDir)
            if not res["OK"]:
                log.debug("Error retrieving directory contents", "%s %s" % (currentDir, res["Message"]))
            else:
                dirContents = res["Value"]
                subdirs = dirContents["SubDirs"]
                files = dirContents["Files"]
                log.debug("%s: %d files, %d sub-directories" % (currentDir, len(files), len(subdirs)))
                for subdir in subdirs:
                    if (not days) or _isOlderThan(subdirs[subdir]["CreationDate"], days):
                        if subdir[0] != "/":
                            subdir = currentDir + "/" + subdir
                        activeDirs.append(subdir)
                for fileName in files:
                    fileInfo = files[fileName]
                    fileInfo = fileInfo.get("Metadata", fileInfo)
                    if (not days) or not fileInfo.get("CreationDate") or _isOlderThan(fileInfo["CreationDate"], days):
                        if wildcard == "*" or fnmatch.fnmatch(fileName, wildcard):
                            fileName = fileInfo.get("LFN", fileName)
                            allFiles.append(fileName)
        return S_OK(allFiles)

    ##########################################################################
    #
    # These are the data transfer methods
    #

    def getFile(self, lfn, destinationDir="", sourceSE=None):
        """Get a local copy of a LFN from Storage Elements.

        'lfn' is the logical file name for the desired file
        """
        log = self.log.getSubLogger("getFile")
        fileMetadata = {}
        if isinstance(lfn, list):
            lfns = lfn
        elif isinstance(lfn, six.string_types):
            lfns = [lfn]
        else:
            errStr = "Supplied lfn must be string or list of strings."
            log.debug(errStr)
            return S_ERROR(errStr)
        log.debug("Attempting to get %s files." % len(lfns))
        res = self.getActiveReplicas(lfns, getUrl=False)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        lfnReplicas = res["Value"]["Successful"]
        # If some files have replicas, check their metadata
        if lfnReplicas:
            res = self.fileCatalog.getFileMetadata(list(lfnReplicas))
            if not res["OK"]:
                return res
            failed.update(res["Value"]["Failed"])
            fileMetadata = res["Value"]["Successful"]
        successful = {}
        for lfn in fileMetadata:
            res = self.__getFile(lfn, lfnReplicas[lfn], fileMetadata[lfn], destinationDir, sourceSE=sourceSE)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            else:
                successful[lfn] = res["Value"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __getFile(self, lfn, replicas, metadata, destinationDir, sourceSE=None):
        """
        Method actually doing the job to get a file from storage
        """

        log = self.log.getSubLogger("__getFile")
        if not replicas:
            errStr = "No accessible replicas found"
            log.debug(errStr)
            return S_ERROR(errStr)
        # Determine the best replicas
        errTuple = ("No SE", "found")
        if sourceSE is None:
            sortedSEs = self._getSEProximity(replicas)
        else:
            if sourceSE not in replicas:
                return S_ERROR("No replica at %s" % sourceSE)
            else:
                sortedSEs = [sourceSE]

        for storageElementName in sortedSEs:
            se = StorageElement(storageElementName, vo=self.voName)

            res = returnSingleResult(se.getFile(lfn, localPath=os.path.realpath(destinationDir)))

            if not res["OK"]:
                errTuple = (
                    "Error getting file from storage:",
                    "%s from %s, %s" % (lfn, storageElementName, res["Message"]),
                )
                errToReturn = res
            else:
                localFile = os.path.realpath(os.path.join(destinationDir, os.path.basename(lfn)))
                localAdler = fileAdler(localFile)

                if metadata["Size"] != res["Value"]:
                    errTuple = (
                        "Mismatch of sizes:",
                        "downloaded = %d, catalog = %d" % (res["Value"], metadata["Size"]),
                    )
                    errToReturn = S_ERROR(DErrno.EFILESIZE, errTuple[1])

                elif (metadata["Checksum"]) and (not compareAdler(metadata["Checksum"], localAdler)):
                    errTuple = (
                        "Mismatch of checksums:",
                        "downloaded = %s, catalog = %s" % (localAdler, metadata["Checksum"]),
                    )
                    errToReturn = S_ERROR(DErrno.EBADCKS, errTuple[1])
                else:
                    return S_OK(localFile)
            # If we are here, there was an error, log it debug level
            log.debug(errTuple[0], errTuple[1])

        log.verbose("Failed to get local copy from any replicas:", "\n%s %s" % errTuple)

        return errToReturn

    def _getSEProximity(self, replicas):
        """get SE proximity"""
        siteName = DIRAC.siteName()
        self.__filterTapeSEs(replicas)
        localSEs = [se for se in self.dmsHelper.getSEsAtSite(siteName).get("Value", []) if se in replicas]
        countrySEs = []
        countryCode = str(siteName).split(".")[-1]
        res = self.dmsHelper.getSEsAtCountry(countryCode)
        if res["OK"]:
            countrySEs = [se for se in res["Value"] if se in replicas and se not in localSEs]
        sortedSEs = randomize(localSEs) + randomize(countrySEs)
        sortedSEs += randomize(se for se in replicas if se not in sortedSEs)

        return sortedSEs

    def putAndRegister(self, lfn, fileName, diracSE, guid=None, path=None, checksum=None, overwrite=False):
        """Put a local file to a Storage Element and register in the File Catalogues

        'lfn' is the file LFN
        'file' is the full path to the local file
        'diracSE' is the Storage Element to which to put the file
        'guid' is the guid with which the file is to be registered (if not provided will be generated)
        'path' is the path on the storage where the file will be put (if not provided the LFN will be used)
        'overwrite' removes file from the file catalogue and SE before attempting upload
        """

        res = self.__hasAccess("addFile", lfn)
        if not res["OK"]:
            return res
        log = self.log.getSubLogger("putAndRegister")
        if lfn not in res["Value"]["Successful"]:
            errStr = "Write access not permitted for this credential."
            log.debug(errStr, lfn)
            return S_ERROR(errStr)

        # Check that the local file exists
        if not os.path.exists(fileName):
            errStr = "Supplied file does not exist."
            log.debug(errStr, fileName)
            return S_ERROR(errStr)
        # If the path is not provided then use the LFN path
        if not path:
            path = os.path.dirname(lfn)
        # Obtain the size of the local file
        size = getSize(fileName)
        if size == 0:
            errStr = "Supplied file is zero size."
            log.debug(errStr, fileName)
            return S_ERROR(errStr)
        # If the GUID is not given, generate it here
        if not guid:
            guid = makeGuid(fileName)
        if not checksum:
            log.debug("Checksum information not provided. Calculating adler32.")
            checksum = fileAdler(fileName)
            # Make another try
            if not checksum:
                log.debug("Checksum calculation failed, try again")
                checksum = fileAdler(fileName)
            if checksum:
                log.debug("Checksum calculated to be %s." % checksum)
            else:
                return S_ERROR(DErrno.EBADCKS, "Unable to calculate checksum")

        res = self.fileCatalog.exists({lfn: guid})
        if not res["OK"]:
            errStr = "Completely failed to determine existence of destination LFN."
            log.debug(errStr, lfn)
            return res
        if lfn not in res["Value"]["Successful"]:
            errStr = "Failed to determine existence of destination LFN."
            log.debug(errStr, lfn)
            return S_ERROR(errStr)
        if res["Value"]["Successful"][lfn]:
            if res["Value"]["Successful"][lfn] == lfn:
                if overwrite:
                    resRm = self.removeFile(lfn, force=True)
                    if not resRm["OK"]:
                        errStr = "Failed to prepare file for overwrite"
                        log.debug(errStr, lfn)
                        return resRm
                    if lfn not in resRm["Value"]["Successful"]:
                        errStr = "Failed to either delete file or LFN"
                        log.debug(errStr, lfn)
                        return S_ERROR("%s %s" % (errStr, lfn))
                else:
                    errStr = "The supplied LFN already exists in the File Catalog."
                    log.debug(errStr, lfn)
                    return S_ERROR("%s %s" % (errStr, res["Value"]["Successful"][lfn]))
            else:
                # If the returned LFN is different, this is the name of a file
                # with the same GUID
                errStr = "This file GUID already exists for another file"
                log.debug(errStr, res["Value"]["Successful"][lfn])
                return S_ERROR("%s %s" % (errStr, res["Value"]["Successful"][lfn]))

        ##########################################################
        #  Instantiate the destination storage element here.
        storageElement = StorageElement(diracSE, vo=self.voName)
        res = storageElement.isValid()
        if not res["OK"]:
            errStr = "The storage element is not currently valid."
            log.verbose(errStr, "%s %s" % (diracSE, res["Message"]))
            return S_ERROR("%s %s" % (errStr, res["Message"]))

        fileDict = {lfn: fileName}

        successful = {}
        failed = {}
        ##########################################################
        #  Perform the put here.
        oDataOperation = _initialiseAccountingObject("putAndRegister", diracSE, 1)
        oDataOperation.setStartTime()
        oDataOperation.setValueByKey("TransferSize", size)
        startTime = time.time()
        res = returnSingleResult(storageElement.putFile(fileDict))
        putTime = time.time() - startTime
        oDataOperation.setValueByKey("TransferTime", putTime)
        if not res["OK"]:

            # We don't consider it a failure if the SE is not valid
            if not DErrno.cmpError(res, errno.EACCES):
                oDataOperation.setValueByKey("TransferOK", 0)
                oDataOperation.setValueByKey("FinalStatus", "Failed")
                oDataOperation.setEndTime()
                gDataStoreClient.addRegister(oDataOperation)
                gDataStoreClient.commit()
                startTime = time.time()
                log.debug("putAndRegister: Sending accounting took %.1f seconds" % (time.time() - startTime))
            errStr = "Failed to put file to Storage Element."
            log.debug(errStr, "%s: %s" % (fileName, res["Message"]))
            return S_ERROR("%s %s" % (errStr, res["Message"]))
        successful[lfn] = {"put": putTime}

        ###########################################################
        # Perform the registration here
        destinationSE = storageElement.storageElementName()
        res = returnSingleResult(storageElement.getURL(lfn, protocol=self.registrationProtocol))
        if not res["OK"]:
            errStr = "Failed to generate destination PFN."
            log.debug(errStr, res["Message"])
            return S_ERROR("%s %s" % (errStr, res["Message"]))
        destUrl = res["Value"]
        oDataOperation.setValueByKey("RegistrationTotal", 1)

        fileTuple = (lfn, destUrl, size, destinationSE, guid, checksum)
        registerDict = {
            "LFN": lfn,
            "PFN": destUrl,
            "Size": size,
            "TargetSE": destinationSE,
            "GUID": guid,
            "Addler": checksum,
        }
        startTime = time.time()
        res = self.registerFile(fileTuple)
        registerTime = time.time() - startTime
        oDataOperation.setValueByKey("RegistrationTime", registerTime)
        if not res["OK"]:
            errStr = "Completely failed to register file."
            log.debug(errStr, res["Message"])
            failed[lfn] = {"register": registerDict}
            oDataOperation.setValueByKey("FinalStatus", "Failed")
        elif lfn in res["Value"]["Failed"]:
            errStr = "Failed to register file."
            log.debug(errStr, "%s %s" % (lfn, res["Value"]["Failed"][lfn]))
            oDataOperation.setValueByKey("FinalStatus", "Failed")
            failed[lfn] = {"register": registerDict}
        else:
            successful[lfn]["register"] = registerTime
            oDataOperation.setValueByKey("RegistrationOK", 1)
        oDataOperation.setEndTime()
        gDataStoreClient.addRegister(oDataOperation)
        startTime = time.time()
        gDataStoreClient.commit()
        log.debug("Sending accounting took %.1f seconds" % (time.time() - startTime))
        return S_OK({"Successful": successful, "Failed": failed})

    def replicateAndRegister(self, lfn, destSE, sourceSE="", destPath="", localCache="", catalog=""):
        """Replicate a LFN to a destination SE and register the replica.

        'lfn' is the LFN to be replicated
        'destSE' is the Storage Element the file should be replicated to
        'sourceSE' is the source for the file replication (where not specified all replicas will be attempted)
        'destPath' is the path on the destination storage element, if to be different from LHCb convention
        'localCache' is the local file system location to be used as a temporary cache
        """
        log = self.log.getSubLogger("replicateAndRegister")
        successful = {}
        failed = {}
        log.debug("Attempting to replicate %s to %s." % (lfn, destSE))
        startReplication = time.time()
        res = self.__replicate(lfn, destSE, sourceSE, destPath, localCache)
        replicationTime = time.time() - startReplication
        if not res["OK"]:
            errStr = "Completely failed to replicate file."
            log.debug(errStr, res["Message"])
            return S_ERROR("%s %s" % (errStr, res["Message"]))
        if not res["Value"]:
            # The file was already present at the destination SE
            log.debug("%s already present at %s." % (lfn, destSE))
            successful[lfn] = {"replicate": 0, "register": 0}
            resDict = {"Successful": successful, "Failed": failed}
            return S_OK(resDict)
        successful[lfn] = {"replicate": replicationTime}

        destPfn = res["Value"]["DestPfn"]
        destSE = res["Value"]["DestSE"]
        log.debug("Attempting to register %s at %s." % (destPfn, destSE))
        replicaTuple = (lfn, destPfn, destSE)
        startRegistration = time.time()
        res = self.registerReplica(replicaTuple, catalog=catalog)
        registrationTime = time.time() - startRegistration
        if not res["OK"]:
            # Need to return to the client that the file was replicated but not
            # registered
            errStr = "Completely failed to register replica."
            log.debug(errStr, res["Message"])
            failed[lfn] = {"Registration": {"LFN": lfn, "TargetSE": destSE, "PFN": destPfn}}
        else:
            if lfn in res["Value"]["Successful"]:
                log.debug("Successfully registered replica.")
                successful[lfn]["register"] = registrationTime
            else:
                errStr = "Failed to register replica."
                log.debug(errStr, res["Value"]["Failed"][lfn])
                failed[lfn] = {"Registration": {"LFN": lfn, "TargetSE": destSE, "PFN": destPfn}}
        return S_OK({"Successful": successful, "Failed": failed})

    def replicate(self, lfn, destSE, sourceSE="", destPath="", localCache=""):
        """Replicate a LFN to a destination SE and register the replica.

        'lfn' is the LFN to be replicated
        'destSE' is the Storage Element the file should be replicated to
        'sourceSE' is the source for the file replication (where not specified all replicas will be attempted)
        'destPath' is the path on the destination storage element, if to be different from LHCb convention
        'localCache' is the local file system location to be used as a temporary cache
        """
        log = self.log.getSubLogger("replicate")
        log.debug("Attempting to replicate %s to %s." % (lfn, destSE))
        res = self.__replicate(lfn, destSE, sourceSE, destPath, localCache)
        if not res["OK"]:
            errStr = "Replication failed."
            log.debug(errStr, "%s %s" % (lfn, destSE))
            return res
        if not res["Value"]:
            # The file was already present at the destination SE
            log.debug("%s already present at %s." % (lfn, destSE))
            return res
        return S_OK(lfn)

    def __getSERealName(self, storageName):
        """get the base name of an SE possibly defined as an alias"""
        rootConfigPath = "/Resources/StorageElements"
        configPath = "%s/%s" % (rootConfigPath, storageName)
        res = gConfig.getOptions(configPath)
        if not res["OK"]:
            errStr = "Failed to get storage options"
            return S_ERROR(errStr)
        if not res["Value"]:
            errStr = "Supplied storage doesn't exist."
            return S_ERROR(errStr)
        if "Alias" in res["Value"]:
            configPath += "/Alias"
            aliasName = gConfig.getValue(configPath)
            result = self.__getSERealName(aliasName)
            if not result["OK"]:
                return result
            resolvedName = result["Value"]
        else:
            resolvedName = storageName
        return S_OK(resolvedName)

    def __isSEInList(self, seName, seList):
        """Check whether an SE is in a list of SEs... All could be aliases"""
        seSet = set()
        for se in seList:
            res = self.__getSERealName(se)
            if res["OK"]:
                seSet.add(res["Value"])
        return self.__getSERealName(seName).get("Value") in seSet

    def __replicate(self, lfn, destSEName, sourceSEName="", destPath="", localCache=""):
        """Replicate a LFN to a destination SE.

        'lfn' is the LFN to be replicated
        'destSE' is the Storage Element the file should be replicated to
        'sourceSE' is the source for the file replication (where not specified all replicas will be attempted)
        'destPath' is the path on the destination storage element, if to be different from LHCb convention
        'localCache' if cannot do third party transfer, we do get and put through this local directory
        """

        log = self.log.getSubLogger("__replicate")

        ###########################################################
        # Check that we have write permissions to this directory.
        res = self.__hasAccess("addReplica", lfn)
        if not res["OK"]:
            return res
        if lfn not in res["Value"]["Successful"]:
            errStr = "__replicate: Write access not permitted for this credential."
            log.debug(errStr, lfn)
            return S_ERROR(errStr)

        # Check that the destination storage element is sane and resolve its name
        log.debug("Verifying destination StorageElement validity (%s)." % (destSEName))

        destStorageElement = StorageElement(destSEName, vo=self.voName)
        res = destStorageElement.isValid()
        if not res["OK"]:
            errStr = "The storage element is not currently valid."
            log.verbose(errStr, "%s %s" % (destSEName, res["Message"]))
            return S_ERROR("%s %s" % (errStr, res["Message"]))

        # Get the real name of the SE
        destSEName = destStorageElement.storageElementName()

        ###########################################################
        # Check whether the destination storage element is banned
        log.verbose("Determining whether %s ( destination ) is Write-banned." % destSEName)

        if not destStorageElement.status()["Write"]:
            infoStr = "Supplied destination Storage Element is not currently allowed for Write."
            log.debug(infoStr, destSEName)
            return S_ERROR(infoStr)

        # Get the LFN replicas from the file catalog
        log.debug("Attempting to obtain replicas for %s." % (lfn))
        res = returnSingleResult(self.getReplicas(lfn, getUrl=False))
        if not res["OK"]:
            errStr = "Failed to get replicas for LFN."
            log.debug(errStr, "%s %s" % (lfn, res["Message"]))
            return S_ERROR("%s %s" % (errStr, res["Message"]))

        log.debug("Successfully obtained replicas for LFN.")
        lfnReplicas = res["Value"]

        ###########################################################
        # If the file catalog size is zero fail the transfer
        log.debug("Attempting to obtain size for %s." % lfn)
        res = returnSingleResult(self.fileCatalog.getFileSize(lfn))
        if not res["OK"]:
            errStr = "Failed to get size for LFN."
            log.debug(errStr, "%s %s" % (lfn, res["Message"]))
            return S_ERROR("%s %s" % (errStr, res["Message"]))

        catalogSize = res["Value"]

        if catalogSize == 0:
            errStr = "Registered file size is 0."
            log.debug(errStr, lfn)
            return S_ERROR(errStr)

        log.debug("File size determined to be %s." % catalogSize)

        ###########################################################
        # If the LFN already exists at the destination we have nothing to do
        if self.__isSEInList(destSEName, lfnReplicas):
            log.debug("__replicate: LFN is already registered at %s." % destSEName)
            return S_OK()

        ###########################################################
        # If the source is specified, check that it is in the replicas

        if sourceSEName:
            log.debug("Determining whether source Storage Element specified is sane.")

            if sourceSEName not in lfnReplicas:
                errStr = "LFN does not exist at supplied source SE."
                log.error(errStr, "%s %s" % (lfn, sourceSEName))
                return S_ERROR(errStr)

        # If sourceSE is specified, then we consider this one only, otherwise
        # we consider them all

        possibleSourceSEs = [sourceSEName] if sourceSEName else lfnReplicas

        # We sort the possibileSourceSEs with the SEs that are on the same site than the destination first
        # reverse = True because True > False
        possibleSourceSEs = sorted(
            possibleSourceSEs,
            key=lambda x: self.dmsHelper.isSameSiteSE(x, destSEName).get("Value", False),
            reverse=True,
        )

        # In case we manage to find SEs that would work as a source, but we can't negotiate a protocol
        # we will do a get and put using one of this sane SE
        possibleIntermediateSEs = []

        # Take into account the destination path
        if destPath:
            destPath = "%s/%s" % (destPath, os.path.basename(lfn))
        else:
            destPath = lfn

        for candidateSEName in possibleSourceSEs:

            log.debug("Consider %s as a source" % candidateSEName)

            # Check that the candidate is active
            if not self.__checkSEStatus(candidateSEName, status="Read"):
                log.debug("%s is currently not allowed as a source." % candidateSEName)
                continue
            else:
                log.debug("%s is available for use." % candidateSEName)

            candidateSE = StorageElement(candidateSEName, vo=self.voName)

            # Check that the SE is valid
            res = candidateSE.isValid()
            if not res["OK"]:
                log.verbose("The storage element is not currently valid.", "%s %s" % (candidateSEName, res["Message"]))
                continue
            else:
                log.debug("The storage is currently valid", candidateSEName)

            # Check that the file size corresponds to the one in the FC
            res = returnSingleResult(candidateSE.getFileSize(lfn))
            if not res["OK"]:
                log.debug("could not get fileSize on %s" % candidateSEName, res["Message"])
                continue
            seFileSize = res["Value"]

            if seFileSize != catalogSize:
                log.debug("Catalog size and physical file size mismatch.", "%s %s" % (catalogSize, seFileSize))
                continue
            else:
                log.debug("Catalog size and physical size match")

            res = destStorageElement.negociateProtocolWithOtherSE(candidateSE, protocols=self.thirdPartyProtocols)

            if not res["OK"]:
                log.debug("Error negotiating replication protocol", res["Message"])
                continue

            replicationProtocols = res["Value"]

            if not replicationProtocols:
                possibleIntermediateSEs.append(candidateSE)
                log.debug("No protocol suitable for replication found")
                continue

            log.debug("Found common protocols", replicationProtocols)

            # THIS WOULD NOT WORK IF PROTO == file !!
            # Why did I write that comment ?!

            # We try the protocols one by one
            # That obviously assumes that there is an overlap and not only
            # a compatibility between the  output protocols of the source
            # and the input protocols of the destination.
            # But that is the only way to make sure we are not replicating
            # over ourselves.
            for compatibleProtocol in replicationProtocols:

                # Compare the urls to make sure we are not overwriting
                res = returnSingleResult(candidateSE.getURL(lfn, protocol=compatibleProtocol))
                if not res["OK"]:
                    log.debug("Cannot get sourceURL", res["Message"])
                    continue

                sourceURL = res["Value"]

                destURL = ""
                res = returnSingleResult(destStorageElement.getURL(destPath, protocol=compatibleProtocol))
                if not res["OK"]:

                    # for some protocols, in particular srm
                    # you might get an error because the file does not exist
                    # which is exactly what we want
                    # in that case, we just keep going with the comparison
                    # since destURL will be an empty string
                    if not DErrno.cmpError(res, errno.ENOENT):
                        log.debug("Cannot get destURL", res["Message"])
                        continue
                    log.debug("File does not exist: Expected error for TargetSE !!")
                else:
                    destURL = res["Value"]

                if sourceURL == destURL:
                    log.debug("Same source and destination, give up")
                    continue

                # Attempt the transfer
                res = returnSingleResult(
                    destStorageElement.replicateFile(
                        {destPath: sourceURL}, sourceSize=catalogSize, inputProtocol=compatibleProtocol
                    )
                )

                if not res["OK"]:
                    log.debug("Replication failed", "%s from %s to %s." % (lfn, candidateSEName, destSEName))
                    continue

                log.debug("Replication successful.", res["Value"])

                res = returnSingleResult(destStorageElement.getURL(destPath, protocol=self.registrationProtocol))
                if not res["OK"]:
                    log.debug("Error getting the registration URL", res["Message"])
                    # it's maybe pointless to try the other candidateSEs...
                    continue

                registrationURL = res["Value"]

                return S_OK({"DestSE": destSEName, "DestPfn": registrationURL})

        # If we are here, that means that we could not make a third party transfer.
        # Check if we have some sane SEs from which we could do a get/put

        localDir = os.path.realpath(localCache if localCache else ".")
        localFile = os.path.join(localDir, os.path.basename(lfn))

        log.debug("Will try intermediate transfer from %s sources" % len(possibleIntermediateSEs))

        for candidateSE in possibleIntermediateSEs:

            res = returnSingleResult(candidateSE.getFile(lfn, localPath=localDir))
            if not res["OK"]:
                log.debug("Error getting the file from %s" % candidateSE.name, res["Message"])
                continue

            res = returnSingleResult(destStorageElement.putFile({destPath: localFile}))

            # Remove the local file whatever happened
            try:
                os.remove(localFile)
            except OSError as e:
                log.error("Error removing local file", "%s %s" % (localFile, e))

            if not res["OK"]:
                log.debug("Error putting file coming from %s" % candidateSE.name, res["Message"])
                # if the put is the problem, it's maybe pointless to try the other
                # candidateSEs...
                continue

            # get URL with default protocol to return it
            res = returnSingleResult(destStorageElement.getURL(destPath, protocol=self.registrationProtocol))
            if not res["OK"]:
                log.debug("Error getting the registration URL", res["Message"])
                # it's maybe pointless to try the other candidateSEs...
                continue

            registrationURL = res["Value"]
            return S_OK({"DestSE": destSEName, "DestPfn": registrationURL})

        # If here, we are really doomed
        errStr = "Failed to replicate with all sources."
        log.debug(errStr, lfn)
        return S_ERROR(errStr)

    ###################################################################
    #
    # These are the file catalog write methods
    #

    def registerFile(self, fileTuple, catalog=""):
        """Register a file or a list of files

        :param self: self reference
        :param tuple fileTuple: (lfn, physicalFile, fileSize, storageElementName, fileGuid, checksum )
        :param str catalog: catalog name
        """
        log = self.log.getSubLogger("registerFile")
        if isinstance(fileTuple, (list, set)):
            fileTuples = fileTuple
        elif isinstance(fileTuple, tuple):
            fileTuples = [fileTuple]
        for fileTuple in fileTuples:
            if not isinstance(fileTuple, tuple):
                errStr = "Supplied file info must be tuple or list of tuples."
                log.debug(errStr)
                return S_ERROR(errStr)
        if not fileTuples:
            return S_OK({"Successful": [], "Failed": {}})
        log.debug("Attempting to register %s files." % len(fileTuples))
        res = self.__registerFile(fileTuples, catalog)
        if not res["OK"]:
            errStr = "Completely failed to register files."
            log.debug(errStr, res["Message"])
            return res
        return res

    def __registerFile(self, fileTuples, catalog):
        """register file to catalog"""

        fileDict = {}

        for lfn, physicalFile, fileSize, storageElementName, fileGuid, checksum in fileTuples:
            fileDict[lfn] = {
                "PFN": physicalFile,
                "Size": fileSize,
                "SE": storageElementName,
                "GUID": fileGuid,
                "Checksum": checksum,
            }

        if catalog:
            fileCatalog = FileCatalog(catalog, vo=self.voName)
            if not fileCatalog.isOK():
                return S_ERROR("Can't get FileCatalog %s" % catalog)
        else:
            fileCatalog = self.fileCatalog

        res = fileCatalog.addFile(fileDict)
        if not res["OK"]:
            errStr = "Completely failed to register files."
            self.log.getSubLogger("__registerFile").debug(errStr, res["Message"])

        return res

    def registerReplica(self, replicaTuple, catalog=""):
        """Register a replica (or list of) supplied in the replicaTuples.

        'replicaTuple' is a tuple or list of tuples of the form (lfn,pfn,se)
        """
        log = self.log.getSubLogger("registerReplica")
        if isinstance(replicaTuple, (list, set)):
            replicaTuples = replicaTuple
        elif isinstance(replicaTuple, tuple):
            replicaTuples = [replicaTuple]
        for replicaTuple in replicaTuples:
            if not isinstance(replicaTuple, tuple):
                errStr = "Supplied file info must be tuple or list of tuples."
                log.debug(errStr)
                return S_ERROR(errStr)
        if not replicaTuples:
            return S_OK({"Successful": [], "Failed": {}})
        log.debug("Attempting to register %s replicas." % len(replicaTuples))
        res = self.__registerReplica(replicaTuples, catalog)
        if not res["OK"]:
            errStr = "Completely failed to register replicas."
            log.debug(errStr, res["Message"])
            return res
        return res

    def __registerReplica(self, replicaTuples, catalog):
        """register replica to catalogue"""
        log = self.log.getSubLogger("__registerReplica")
        seDict = {}
        for lfn, url, storageElementName in replicaTuples:
            seDict.setdefault(storageElementName, []).append((lfn, url))
        failed = {}
        replicaTuples = []
        for storageElementName, replicaTuple in seDict.items():  # can be an iterator
            destStorageElement = StorageElement(storageElementName, vo=self.voName)
            res = destStorageElement.isValid()
            if not res["OK"]:
                errStr = "The storage element is not currently valid."
                log.verbose(errStr, "%s %s" % (storageElementName, res["Message"]))
                for lfn, url in replicaTuple:
                    failed[lfn] = errStr
            else:
                storageElementName = destStorageElement.storageElementName()
                for lfn, url in replicaTuple:
                    res = returnSingleResult(destStorageElement.getURL(lfn, protocol=self.registrationProtocol))
                    if not res["OK"]:
                        failed[lfn] = res["Message"]
                    else:
                        replicaTuple = (lfn, res["Value"], storageElementName, False)
                        replicaTuples.append(replicaTuple)
        log.debug("Successfully resolved %s replicas for registration." % len(replicaTuples))
        # HACK!
        replicaDict = {}
        for lfn, url, se, _master in replicaTuples:
            replicaDict[lfn] = {"SE": se, "PFN": url}

        if catalog:
            fileCatalog = FileCatalog(catalog, vo=self.voName)
            res = fileCatalog.addReplica(replicaDict)
        else:
            res = self.fileCatalog.addReplica(replicaDict)
        if not res["OK"]:
            errStr = "Completely failed to register replicas."
            log.debug(errStr, res["Message"])
            return S_ERROR("%s %s" % (errStr, res["Message"]))
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        resDict = {"Successful": successful, "Failed": failed}
        return S_OK(resDict)

    ###################################################################
    #
    # These are the removal methods for physical and catalogue removal
    #

    def removeFile(self, lfn, force=None):
        """Remove the file (all replicas) from Storage Elements and file catalogue

        'lfn' is the file to be removed
        """
        log = self.log.getSubLogger("removeFile")
        if not lfn:
            return S_OK({"Successful": {}, "Failed": {}})
        if force is None:
            force = self.ignoreMissingInFC
        if isinstance(lfn, (list, dict, set, tuple)):
            lfns = list(lfn)
        else:
            lfns = [lfn]
        for lfn in lfns:
            if not isinstance(lfn, six.string_types):
                errStr = "Supplied lfns must be string or list of strings."
                log.debug(errStr)
                return S_ERROR(errStr)

        successful = {}
        failed = {}
        if not lfns:
            return S_OK({"Successful": successful, "Failed": failed})
        # First check if the file exists in the FC
        res = self.fileCatalog.exists(lfns)
        if not res["OK"]:
            return res
        success = res["Value"]["Successful"]
        lfns = [lfn for lfn in success if success[lfn]]
        if force:
            # Files that don't exist are removed successfully
            successful = dict.fromkeys((lfn for lfn in success if not success[lfn]), True)
        else:
            failed = dict.fromkeys((lfn for lfn in success if not success[lfn]), "No such file or directory")
        # Check that we have write permissions to this directory and to the file.
        if lfns:
            res = self.__hasAccess("removeFile", lfns)
            if not res["OK"]:
                return res
            if res["Value"]["Failed"]:
                errStr = "Write access not permitted for this credential."
                log.debug(errStr, "for %d files" % len(res["Value"]["Failed"]))
                failed.update(dict.fromkeys(res["Value"]["Failed"], errStr))

            lfns = res["Value"]["Successful"]

            if lfns:
                log.debug("Attempting to remove %d files from Storage and Catalogue. Get replicas first" % len(lfns))
                res = self.fileCatalog.getReplicas(lfns, allStatus=True)
                if not res["OK"]:
                    errStr = "DataManager.removeFile: Completely failed to get replicas for lfns."
                    log.debug(errStr, res["Message"])
                    return res
                lfnDict = res["Value"]["Successful"]

                for lfn, reason in res["Value"]["Failed"].items():  # can be an iterator
                    # Ignore files missing in FC if force is set
                    if reason == "No such file or directory" and force:
                        successful[lfn] = True
                    elif reason == "File has zero replicas":
                        lfnDict[lfn] = {}
                    else:
                        failed[lfn] = reason

                res = self.__removeFile(lfnDict)
                if not res["OK"]:
                    # This can never happen
                    return res
                failed.update(res["Value"]["Failed"])
                successful.update(res["Value"]["Successful"])

        gDataStoreClient.commit()
        return S_OK({"Successful": successful, "Failed": failed})

    def __removeFile(self, lfnDict):
        """remove file"""
        storageElementDict = {}
        # # sorted and reversed
        for lfn, repDict in sorted(lfnDict.items(), reverse=True):
            for se in repDict:
                storageElementDict.setdefault(se, []).append(lfn)
        failed = {}
        successful = {}
        for storageElementName in sorted(storageElementDict):
            lfns = storageElementDict[storageElementName]
            res = self.__removeReplica(storageElementName, lfns, replicaDict=lfnDict)
            if not res["OK"]:
                errStr = res["Message"]
                for lfn in lfns:
                    failed[lfn] = failed.setdefault(lfn, "") + " %s" % errStr
            else:
                for lfn, errStr in res["Value"]["Failed"].items():  # can be an iterator
                    failed[lfn] = failed.setdefault(lfn, "") + " %s" % errStr

        completelyRemovedFiles = set(lfnDict) - set(failed)
        if completelyRemovedFiles:
            res = self.fileCatalog.removeFile(list(completelyRemovedFiles))
            if not res["OK"]:
                failed.update(
                    dict.fromkeys(completelyRemovedFiles, "Failed to remove file from the catalog: %s" % res["Message"])
                )
            else:
                failed.update(res["Value"]["Failed"])
                successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def removeReplica(self, storageElementName, lfn):
        """Remove replica at the supplied Storage Element from Storage Element then file catalogue

        'storageElementName' is the storage where the file is to be removed
        'lfn' is the file to be removed
        """
        log = self.log.getSubLogger("removeReplica")
        if isinstance(lfn, (list, dict, set, tuple)):
            lfns = set(lfn)
        else:
            lfns = set([lfn])
        for lfn in lfns:
            if not isinstance(lfn, six.string_types):
                errStr = "Supplied lfns must be string or list of strings."
                log.debug(errStr)
                return S_ERROR(errStr)

        successful = {}
        failed = {}
        if not lfns:
            return S_OK({"Successful": successful, "Failed": failed})

        # Check that we have write permissions to this file.
        res = self.__hasAccess("removeReplica", lfns)
        if not res["OK"]:
            log.debug("Error in __verifyWritePermisison", res["Message"])
            return res
        if res["Value"]["Failed"]:
            errStr = "Write access not permitted for this credential."
            log.debug(errStr, "for %d files" % len(res["Value"]["Failed"]))
            failed.update(dict.fromkeys(res["Value"]["Failed"], errStr))
            lfns -= set(res["Value"]["Failed"])

        if not lfns:
            log.debug("Permission denied for all files")
        else:
            log.debug("Will remove %s lfns at %s." % (len(lfns), storageElementName))
            res = self.fileCatalog.getReplicas(list(lfns), allStatus=True)
            if not res["OK"]:
                errStr = "Completely failed to get replicas for lfns."
                log.debug(errStr, res["Message"])
                return res
            failed.update(res["Value"]["Failed"])
            replicaDict = res["Value"]["Successful"]
            lfnsToRemove = set()
            for lfn, repDict in replicaDict.items():  # can be an iterator
                if storageElementName not in repDict:
                    # The file doesn't exist at the storage element so don't have to
                    # remove it
                    successful[lfn] = True
                elif len(repDict) == 1:
                    # The file has only a single replica so don't remove
                    log.debug(
                        "The replica you are trying to remove is the only one.", "%s @ %s" % (lfn, storageElementName)
                    )
                    failed[lfn] = "Failed to remove sole replica"
                else:
                    lfnsToRemove.add(lfn)
            if lfnsToRemove:
                res = self.__removeReplica(storageElementName, lfnsToRemove, replicaDict=replicaDict)
                if not res["OK"]:
                    log.debug("Failed in __removeReplica", res["Message"])
                    return res
                failed.update(res["Value"]["Failed"])
                successful.update(res["Value"]["Successful"])
                gDataStoreClient.commit()
        return S_OK({"Successful": successful, "Failed": failed})

    def __removeReplica(self, storageElementName, lfns, replicaDict=None):
        """remove replica
        Remove the replica from the storageElement, and then from the catalog

        :param storageElementName : The name of the storage Element
        :param lfns : list of lfn we want to remove
        :param replicaDict : cache of fc.getReplicas(lfns) : { lfn { se : catalog url } }

        """
        log = self.log.getSubLogger("__removeReplica")
        failed = {}
        successful = {}
        replicaDict = replicaDict if replicaDict else {}

        lfnsToRemove = set()
        for lfn in lfns:
            res = self.__hasAccess("removeReplica", lfn)
            if not res["OK"]:
                log.debug("Error in __verifyWritePermission", res["Message"])
                return res
            if lfn not in res["Value"]["Successful"]:
                errStr = "Write access not permitted for this credential."
                log.debug(errStr, lfn)
                failed[lfn] = errStr
            else:
                lfnsToRemove.add(lfn)

        # Remove physical replicas first
        res = self.__removePhysicalReplica(storageElementName, lfnsToRemove, replicaDict=replicaDict)
        if not res["OK"]:
            errStr = "Failed to remove physical replicas."
            log.debug(errStr, res["Message"])
            return res
        failed.update(res["Value"]["Failed"])

        # Here we use the FC PFN...
        replicaTuples = [
            (lfn, replicaDict[lfn][storageElementName], storageElementName) for lfn in res["Value"]["Successful"]
        ]
        if replicaTuples:
            res = self.__removeCatalogReplica(replicaTuples)
            if not res["OK"]:
                errStr = "Completely failed to remove physical files."
                log.debug(errStr, res["Message"])
                failed.update(dict.fromkeys((lfn for lfn, _pfn, _se in replicaTuples), res["Message"]))
                successful = {}
            else:
                failed.update(res["Value"]["Failed"])
                successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def removeReplicaFromCatalog(self, storageElementName, lfn):
        """remove :lfn: replica from :storageElementName: SE

        :param self: self reference
        :param str storageElementName: SE name
        :param mixed lfn: a single LFN or list of LFNs
        """
        # FIXME: this method is dangerous ans should eventually be removed as well
        # as the script dirac-dms-remove-catalog-replicas
        log = self.log.getSubLogger("removeReplicaFromCatalog")
        # Remove replica from the file catalog 'lfn' are the file
        # to be removed 'storageElementName' is the storage where the file is to
        # be removed
        if isinstance(lfn, (list, dict, set, tuple)):
            lfns = list(lfn)
        else:
            lfns = [lfn]
        for lfn in lfns:
            if not isinstance(lfn, six.string_types):
                errStr = "Supplied lfns must be string or list of strings."
                log.debug(errStr)
                return S_ERROR(errStr)
        successful = {}
        failed = {}
        if not lfns:
            return S_OK({"Successful": successful, "Failed": failed})
        log.debug("Will remove catalogue entry for %s lfns at %s." % (len(lfns), storageElementName))
        res = self.fileCatalog.getReplicas(lfns, allStatus=True)
        if not res["OK"]:
            errStr = "Completely failed to get replicas for lfns."
            log.debug(errStr, res["Message"])
            return res
        failed = {}
        successful = {}
        for lfn, reason in res["Value"]["Failed"].items():  # can be an iterator
            if reason in ("No such file or directory", "File has zero replicas"):
                successful[lfn] = True
            else:
                failed[lfn] = reason
        replicaTuples = []
        for lfn, repDict in res["Value"]["Successful"].items():  # can be an iterator
            if storageElementName not in repDict:
                # The file doesn't exist at the storage element so don't have to remove
                # it
                successful[lfn] = True
            else:
                replicaTuples.append((lfn, repDict[storageElementName], storageElementName))
        log.debug("Resolved %s pfns for catalog removal at %s." % (len(replicaTuples), storageElementName))
        res = self.__removeCatalogReplica(replicaTuples)
        failed.update(res["Value"]["Failed"])
        successful.update(res["Value"]["Successful"])
        resDict = {"Successful": successful, "Failed": failed}
        return S_OK(resDict)

    def __removeCatalogReplica(self, replicaTuples):
        """remove replica form catalogue
        :param replicaTuples : list of (lfn, catalogPFN, se)
        """
        log = self.log.getSubLogger("__removeCatalogReplica")
        oDataOperation = _initialiseAccountingObject("removeCatalogReplica", "", len(replicaTuples))
        oDataOperation.setStartTime()
        start = time.time()
        # HACK!
        replicaDict = {}
        for lfn, pfn, se in replicaTuples:
            replicaDict[lfn] = {"SE": se, "PFN": pfn}
        res = self.fileCatalog.removeReplica(replicaDict)
        oDataOperation.setEndTime()
        oDataOperation.setValueByKey("RegistrationTime", time.time() - start)
        if not res["OK"]:
            oDataOperation.setValueByKey("RegistrationOK", 0)
            oDataOperation.setValueByKey("FinalStatus", "Failed")
            gDataStoreClient.addRegister(oDataOperation)
            errStr = "Completely failed to remove replica: "
            log.debug(errStr, res["Message"])
            return S_ERROR("%s %s" % (errStr, res["Message"]))

        success = res["Value"]["Successful"]
        failed = res["Value"]["Failed"]
        for lfn, error in list(failed.items()):
            # Ignore error if file doesn't exist
            # This assumes all catalogs return an error as { catalog : error }
            for catalog, err in list(error.items()):
                if "no such file" in err.lower():
                    success.setdefault(lfn, {}).update({catalog: True})
                    error.pop(catalog)
            if not failed[lfn]:
                failed.pop(lfn)
            else:
                log.error("Failed to remove replica.", "%s %s" % (lfn, error))

        # Only for logging information
        if success:
            log.debug("Removed %d replicas" % len(success))
            for lfn in success:
                log.debug("Successfully removed replica.", lfn)

        oDataOperation.setValueByKey("RegistrationOK", len(success))
        gDataStoreClient.addRegister(oDataOperation)
        return res

    def __removePhysicalReplica(self, storageElementName, lfnsToRemove, replicaDict=None):
        """remove replica from storage element

        :param storageElementName : name of the storage Element
        :param lfnsToRemove : set of lfn to removes
        :param replicaDict : cache of fc.getReplicas, to be passed to the SE
        """
        log = self.log.getSubLogger("__removePhysicalReplica")
        log.debug("Attempting to remove %s pfns at %s." % (len(lfnsToRemove), storageElementName))
        storageElement = StorageElement(storageElementName, vo=self.voName)
        res = storageElement.isValid()
        if not res["OK"]:
            errStr = "The storage element is not currently valid."
            log.verbose(errStr, "%s %s" % (storageElementName, res["Message"]))
            return S_ERROR("%s %s" % (errStr, res["Message"]))
        oDataOperation = _initialiseAccountingObject("removePhysicalReplica", storageElementName, len(lfnsToRemove))
        oDataOperation.setStartTime()
        start = time.time()
        lfnsToRemove = list(lfnsToRemove)
        ret = storageElement.getFileSize(lfnsToRemove, replicaDict=replicaDict)
        deletedSizes = ret.get("Value", {}).get("Successful", {})
        res = storageElement.removeFile(lfnsToRemove, replicaDict=replicaDict)
        oDataOperation.setEndTime()
        oDataOperation.setValueByKey("TransferTime", time.time() - start)
        if not res["OK"]:
            oDataOperation.setValueByKey("TransferOK", 0)
            oDataOperation.setValueByKey("FinalStatus", "Failed")
            gDataStoreClient.addRegister(oDataOperation)
            log.debug("Failed to remove replicas.", res["Message"])
        else:
            for lfn, value in list(res["Value"]["Failed"].items()):
                if "No such file or directory" in value:
                    res["Value"]["Successful"][lfn] = lfn
                    res["Value"]["Failed"].pop(lfn)
            for lfn in res["Value"]["Successful"]:
                res["Value"]["Successful"][lfn] = True

            deletedSize = sum(deletedSizes.get(lfn, 0) for lfn in res["Value"]["Successful"])
            oDataOperation.setValueByKey("TransferSize", deletedSize)
            oDataOperation.setValueByKey("TransferOK", len(res["Value"]["Successful"]))

            gDataStoreClient.addRegister(oDataOperation)
            infoStr = "Successfully issued accounting removal request."
            log.debug(infoStr)
        return res

    #########################################################################
    #
    # File transfer methods
    #

    def put(self, lfn, fileName, diracSE, path=None):
        """Put a local file to a Storage Element

        :param self: self reference
        :param str lfn: LFN
        :param str fileName: the full path to the local file
        :param str diracSE: the Storage Element to which to put the file
        :param str path: the path on the storage where the file will be put (if not provided the LFN will be used)

        """
        log = self.log.getSubLogger("put")
        # Check that the local file exists
        if not os.path.exists(fileName):
            errStr = "Supplied file does not exist."
            log.debug(errStr, fileName)
            return S_ERROR(errStr)
        # If the path is not provided then use the LFN path
        if not path:
            path = os.path.dirname(lfn)
        # Obtain the size of the local file
        size = getSize(fileName)
        if size == 0:
            errStr = "Supplied file is zero size."
            log.debug(errStr, fileName)
            return S_ERROR(errStr)

        ##########################################################
        #  Instantiate the destination storage element here.
        storageElement = StorageElement(diracSE, vo=self.voName)
        res = storageElement.isValid()
        if not res["OK"]:
            errStr = "The storage element is not currently valid."
            log.verbose(errStr, "%s %s" % (diracSE, res["Message"]))
            return S_ERROR("%s %s" % (errStr, res["Message"]))
        fileDict = {lfn: fileName}

        successful = {}
        failed = {}
        ##########################################################
        #  Perform the put here.
        startTime = time.time()
        res = returnSingleResult(storageElement.putFile(fileDict))
        putTime = time.time() - startTime
        if not res["OK"]:
            errStr = "Failed to put file to Storage Element."
            failed[lfn] = res["Message"]
            log.debug(errStr, "%s: %s" % (fileName, res["Message"]))
        else:
            log.debug("Put file to storage in %s seconds." % putTime)
            successful[lfn] = res["Value"]
        resDict = {"Successful": successful, "Failed": failed}
        return S_OK(resDict)

    #########################################################################
    #
    # File catalog methods
    #

    def getActiveReplicas(self, lfns, getUrl=True, diskOnly=False, preferDisk=False):
        """Get all the replicas for the SEs which are in Active status for reading."""
        return self.getReplicas(
            lfns, allStatus=False, getUrl=getUrl, diskOnly=diskOnly, preferDisk=preferDisk, active=True
        )

    def __filterTapeReplicas(self, replicaDict, diskOnly=False):
        """
        Check a replica dictionary for disk replicas:
        If there is a disk replica, removetape replicas, else keep all
        The input argument is modified
        """
        seList = set(se for ses in replicaDict["Successful"].values() for se in ses)  # can be an iterator
        # Get a cache of SE statuses for long list of replicas
        seStatus = dict(
            (se, (self.__checkSEStatus(se, status="DiskSE"), self.__checkSEStatus(se, status="TapeSE")))
            for se in seList
        )
        # Beware, there is a del below
        for lfn, replicas in list(replicaDict["Successful"].items()):
            self.__filterTapeSEs(replicas, diskOnly=diskOnly, seStatus=seStatus)
            # If diskOnly, one may not have any replica in the end, set Failed
            if diskOnly and not replicas:
                del replicaDict["Successful"][lfn]
                replicaDict["Failed"][lfn] = "No disk replicas"
        return

    def __filterReplicasForJobs(self, replicaDict):
        """Remove the SEs that are not to be used for jobs, and archive SEs if there are others
        The input argument is modified
        """
        seList = set(se for ses in replicaDict["Successful"].values() for se in ses)  # can be an iterator
        # Get a cache of SE statuses for long list of replicas
        seStatus = dict((se, (self.dmsHelper.isSEForJobs(se), self.dmsHelper.isSEArchive(se))) for se in seList)
        # Beware, there is a del below
        for lfn, replicas in list(replicaDict["Successful"].items()):
            otherThanArchive = set(se for se in replicas if not seStatus[se][1])
            for se in list(replicas):
                # Remove the SE if it should not be used for jobs or if it is an
                # archive and there are other SEs
                if not seStatus[se][0] or (otherThanArchive and seStatus[se][1]):
                    replicas.pop(se)
            # If in the end there is no replica, set Failed
            if not replicas:
                del replicaDict["Successful"][lfn]
                replicaDict["Failed"][lfn] = "No replicas for jobs"
        return

    def __filterTapeSEs(self, replicas, diskOnly=False, seStatus=None):
        """Remove the tape SEs as soon as there is one disk SE or diskOnly is requested
        The input argument is modified
        """
        # Build the SE status cache if not existing
        if seStatus is None:
            seStatus = dict(
                (se, (self.__checkSEStatus(se, status="DiskSE"), self.__checkSEStatus(se, status="TapeSE")))
                for se in replicas
            )

        for se in replicas:  # There is a del below but we then return!
            # First find a disk replica, otherwise do nothing unless diskOnly is set
            if diskOnly or seStatus[se][0]:
                # There is one disk replica, remove tape replicas and exit loop
                for se in list(replicas):  # Beware: there is a pop below
                    if seStatus[se][1]:
                        replicas.pop(se)
                return
        return

    def checkActiveReplicas(self, replicaDict):
        """
        Check a replica dictionary for active replicas, and verify input structure first
        """
        if not isinstance(replicaDict, dict):
            return S_ERROR("Wrong argument type %s, expected a dictionary" % type(replicaDict))

        for key in ["Successful", "Failed"]:
            if key not in replicaDict:
                return S_ERROR('Missing key "%s" in replica dictionary' % key)
            if not isinstance(replicaDict[key], dict):
                return S_ERROR("Wrong argument type %s, expected a dictionary" % type(replicaDict[key]))

        activeDict = {"Successful": {}, "Failed": replicaDict["Failed"].copy()}
        for lfn, replicas in replicaDict["Successful"].items():  # can be an iterator
            if not isinstance(replicas, dict):
                activeDict["Failed"][lfn] = "Wrong replica info"
            else:
                activeDict["Successful"][lfn] = replicas.copy()
        self.__filterActiveReplicas(activeDict)
        return S_OK(activeDict)

    def __filterActiveReplicas(self, replicaDict):
        """
        Check a replica dictionary for active replicas
        The input dict is modified, no returned value
        """
        seList = set(se for ses in replicaDict["Successful"].values() for se in ses)  # can be an iterator
        # Get a cache of SE statuses for long list of replicas
        seStatus = dict((se, self.__checkSEStatus(se, status="Read")) for se in seList)
        for replicas in replicaDict["Successful"].values():  # can be an iterator
            for se in list(replicas):  # Beware: there is a pop below
                if not seStatus[se]:
                    replicas.pop(se)
        return

    def __checkSEStatus(self, se, status="Read"):
        """returns the value of a certain SE status flag (access or other)"""
        return StorageElement(se, vo=self.voName).status().get(status, False)

    def getReplicas(self, lfns, allStatus=True, getUrl=True, diskOnly=False, preferDisk=False, active=False):
        """get replicas from catalogue and filter if requested
        Warning: all filters are independent, hence active and preferDisk should be set if using forJobs
        """
        catalogReplicas = {}
        failed = {}
        for lfnChunk in breakListIntoChunks(lfns, 1000):
            res = self.fileCatalog.getReplicas(lfnChunk, allStatus=allStatus)
            if res["OK"]:
                catalogReplicas.update(res["Value"]["Successful"])
                failed.update(res["Value"]["Failed"])
            else:
                return res
        if not getUrl:
            for lfn in catalogReplicas:
                catalogReplicas[lfn] = dict.fromkeys(catalogReplicas[lfn], True)
        elif not self.useCatalogPFN:
            se_lfn = {}

            # We group the query to getURL by storage element to gain in speed
            for lfn in catalogReplicas:
                for se in catalogReplicas[lfn]:
                    se_lfn.setdefault(se, []).append(lfn)

            for se in se_lfn:
                seObj = StorageElement(se, vo=self.voName)
                succPfn = (
                    seObj.getURL(se_lfn[se], protocol=self.registrationProtocol).get("Value", {}).get("Successful", {})
                )
                for lfn in succPfn:
                    catalogReplicas[lfn][se] = succPfn[lfn]

        result = {"Successful": catalogReplicas, "Failed": failed}
        if active:
            self.__filterActiveReplicas(result)
        if diskOnly or preferDisk:
            self.__filterTapeReplicas(result, diskOnly=diskOnly)
        return S_OK(result)

    def getReplicasForJobs(self, lfns, allStatus=False, getUrl=True, diskOnly=False):
        """get replicas useful for jobs"""
        # Call getReplicas with no filter and enforce filters in this method
        result = self.getReplicas(lfns, allStatus=allStatus, getUrl=getUrl)
        if not result["OK"]:
            return result
        replicaDict = result["Value"]
        # For jobs replicas must be active
        self.__filterActiveReplicas(replicaDict)
        # For jobs, give preference to disk replicas but not only
        self.__filterTapeReplicas(replicaDict, diskOnly=diskOnly)
        # don't use SEs excluded for jobs (e.g. Failover)
        self.__filterReplicasForJobs(replicaDict)
        return S_OK(replicaDict)

    # 3
    # Methods from the catalogToStorage. It would all work with the direct call to the SE, but this checks
    # first if the replica is known to the catalog

    def __executeIfReplicaExists(self, storageElementName, lfn, method, **kwargs):
        """a simple wrapper that allows replica querying then perform the StorageElement operation

        :param self: self reference
        :param str storageElementName: DIRAC SE name
        :param mixed lfn: a LFN str, list of LFNs or dict with LFNs as keys
        """
        log = self.log.getSubLogger("__executeIfReplicaExists")
        # # default value
        kwargs = kwargs if kwargs else {}
        # # get replicas for lfn
        res = FileCatalog(vo=self.voName).getReplicas(lfn)
        if not res["OK"]:
            errStr = "Completely failed to get replicas for LFNs."
            log.debug(errStr, res["Message"])
            return res
        # # returned dict, get failed replicase
        retDict = {"Failed": res["Value"]["Failed"], "Successful": {}}
        # # print errors
        for lfn, reason in retDict["Failed"].items():  # can be an iterator
            log.error("_callReplicaSEFcn: Failed to get replicas for file.", "%s %s" % (lfn, reason))
        # # good replicas
        lfnReplicas = res["Value"]["Successful"]
        # # store PFN to LFN mapping
        lfnList = []
        for lfn, replicas in lfnReplicas.items():  # can be an iterator
            if storageElementName in replicas:
                lfnList.append(lfn)
            else:
                errStr = "File hasn't got replica at supplied Storage Element."
                log.error(errStr, "%s %s" % (lfn, storageElementName))
                retDict["Failed"][lfn] = errStr

        if "replicaDict" not in kwargs:
            kwargs["replicaDict"] = lfnReplicas

        # # call StorageElement function at least
        se = StorageElement(storageElementName, vo=self.voName)
        fcn = getattr(se, method)
        res = fcn(lfnList, **kwargs)
        # # check result
        if not res["OK"]:
            errStr = "Failed to execute %s StorageElement method." % method
            log.error(errStr, res["Message"])
            return res

        # # filter out failed and successful
        retDict["Successful"].update(res["Value"]["Successful"])
        retDict["Failed"].update(res["Value"]["Failed"])

        return S_OK(retDict)

    def getReplicaIsFile(self, lfn, storageElementName):
        """determine whether the supplied lfns are files at the supplied StorageElement

        :param self: self reference
        :param mixed lfn: LFN string, list if LFNs or dict with LFNs as keys
        :param str storageElementName: DIRAC SE name
        :param bool singleFile: execute for the first LFN only
        """
        return self.__executeIfReplicaExists(storageElementName, lfn, "isFile")

    def getReplicaSize(self, lfn, storageElementName):
        """get the size of files for the lfns at the supplied StorageElement

        :param self: self reference
        :param mixed lfn: LFN string, list if LFNs or dict with LFNs as keys
        :param str storageElementName: DIRAC SE name
        :param bool singleFile: execute for the first LFN only
        """
        return self.__executeIfReplicaExists(storageElementName, lfn, "getFileSize")

    def getReplicaAccessUrl(self, lfn, storageElementName, protocol=False):
        """get the access url for lfns at the supplied StorageElement

        :param self: self reference
        :param mixed lfn: LFN string, list if LFNs or dict with LFNs as keys
        :param str storageElementName: DIRAC SE name
        :param bool singleFile: execute for the first LFN only
        """
        return self.__executeIfReplicaExists(storageElementName, lfn, "getURL", protocol=protocol)

    def getReplicaMetadata(self, lfn, storageElementName):
        """get the file metadata for lfns at the supplied StorageElement

        :param self: self reference
        :param mixed lfn: LFN string, list if LFNs or dict with LFNs as keys
        :param str storageElementName: DIRAC SE name
        :param bool singleFile: execute for the first LFN only
        """
        return self.__executeIfReplicaExists(storageElementName, lfn, "getFileMetadata")

    def prestageReplica(self, lfn, storageElementName, lifetime=86400):
        """issue a prestage requests for the lfns at the supplied StorageElement

        :param self: self reference
        :param mixed lfn: LFN string, list if LFNs or dict with LFNs as keys
        :param str storageElementName: DIRAC SE name
        :param int lifetime: 24h in seconds
        :param bool singleFile: execute for the first LFN only
        """
        return self.__executeIfReplicaExists(storageElementName, lfn, "prestageFile", lifetime=lifetime)

    def pinReplica(self, lfn, storageElementName, lifetime=86400):
        """pin the lfns at the supplied StorageElement

        :param self: self reference
        :param mixed lfn: LFN string, list if LFNs or dict with LFNs as keys
        :param str storageElementName: DIRAC SE name
        :param int lifetime: 24h in seconds
        :param bool singleFile: execute for the first LFN only
        """
        return self.__executeIfReplicaExists(storageElementName, lfn, "pinFile", lifetime=lifetime)

    def releaseReplica(self, lfn, storageElementName):
        """release pins for the lfns at the supplied StorageElement

        :param self: self reference
        :param mixed lfn: LFN string, list if LFNs or dict with LFNs as keys
        :param str storageElementName: DIRAC SE name
        :param bool singleFile: execute for the first LFN only
        """
        return self.__executeIfReplicaExists(storageElementName, lfn, "releaseFile")

    def getReplica(self, lfn, storageElementName, localPath=False):
        """copy replicas from DIRAC SE to local directory

        :param self: self reference
        :param mixed lfn: LFN string, list if LFNs or dict with LFNs as keys
        :param str storageElementName: DIRAC SE name
        :param mixed localPath: path in the local file system, if False, os.getcwd() will be used
        :param bool singleFile: execute for the first LFN only
        """
        return self.__executeIfReplicaExists(storageElementName, lfn, "getFile", localPath=localPath)



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-site-mask-logging
# Author :  Stuart Paterson
########################################################################
"""
Retrieves site mask logging information.

Example:
  $ dirac-admin-site-mask-logging LCG.IN2P3.fr
  Site Mask Logging Info for LCG.IN2P3.fr
  Active  2010-12-08 21:28:16 ( atsareg )
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["Site:     Name of the Site"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []

    for site in args:
        result = diracAdmin.getSiteMaskLogging(site, printOutput=True)
        if not result["OK"]:
            errorList.append((site, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



"""
This module contains constants and lists for the possible pilot states.
"""
__RCSID__ = "$Id$"

#: The pilot has been generated and is transferred to a remote site:
SUBMITTED = "Submitted"
#: The pilot is waiting for a computing resource in a batch queue:
WAITING = "Waiting"
#: The pilot is running a payload on a worker node:
RUNNING = "Running"
#: The pilot finished its execution:
DONE = "Done"
#: The pilot execution failed:
FAILED = "Failed"
#: The pilot was deleted:
DELETED = "Deleted"
#: The pilot execution was aborted:
ABORTED = "Aborted"
#: Cannot get information about the pilot status:
UNKNOWN = "Unknown"

# Note: 'Scheduled' is deprecated and should disappear in future releases
#: Possible pilot states:
PILOT_STATES = [SUBMITTED, WAITING, RUNNING, DONE, FAILED, DELETED, ABORTED, UNKNOWN, "Scheduled"]

#: Waiting states:
PILOT_WAITING_STATES = [SUBMITTED, WAITING, "Scheduled"]

#: Transient states:
PILOT_TRANSIENT_STATES = list(PILOT_WAITING_STATES + [RUNNING, UNKNOWN])


#: Final states:
PILOT_FINAL_STATES = [DONE, ABORTED, DELETED, FAILED]



"""StageMonitorAgent

This agents queries the storage element about staging requests, to see if files are staged or not.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN StageMonitorAgent
  :end-before: ##END
  :caption: StageMonitorAgent options
  :dedent: 2

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK, S_ERROR, siteName

from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.AccountingSystem.Client.Types.DataOperation import DataOperation
from DIRAC.AccountingSystem.Client.DataStoreClient import gDataStoreClient
from DIRAC.Core.Security.ProxyInfo import getProxyInfo

import re

AGENT_NAME = "StorageManagement/StageMonitorAgent"


class StageMonitorAgent(AgentModule):
    def initialize(self):
        self.stagerClient = StorageManagerClient()
        # This sets the Default Proxy to used as that defined under
        # /Operations/Shifter/DataManager
        # the shifterProxy option in the Configuration can be used to change this default.
        self.am_setOption("shifterProxy", "DataManager")
        self.storagePlugins = self.am_getOption("StoragePlugins", [])

        return S_OK()

    def execute(self):

        res = getProxyInfo(disableVOMS=True)
        if not res["OK"]:
            return res
        self.proxyInfoDict = res["Value"]

        return self.monitorStageRequests()

    def monitorStageRequests(self):
        """This is the third logical task manages the StageSubmitted->Staged transition of the Replicas"""
        res = self.__getStageSubmittedReplicas()
        if not res["OK"]:
            gLogger.fatal(
                "StageMonitor.monitorStageRequests: Failed to get replicas from StorageManagementDB.", res["Message"]
            )
            return res
        if not res["Value"]:
            gLogger.info("StageMonitor.monitorStageRequests: There were no StageSubmitted replicas found")
            return res
        seReplicas = res["Value"]["SEReplicas"]
        replicaIDs = res["Value"]["ReplicaIDs"]
        gLogger.info(
            "StageMonitor.monitorStageRequests: Obtained %s StageSubmitted replicas for monitoring." % len(replicaIDs)
        )
        for storageElement, seReplicaIDs in seReplicas.items():
            self.__monitorStorageElementStageRequests(storageElement, seReplicaIDs, replicaIDs)

        gDataStoreClient.commit()

        return S_OK()

    def __monitorStorageElementStageRequests(self, storageElement, seReplicaIDs, replicaIDs):
        terminalReplicaIDs = {}
        oldRequests = []
        stagedReplicas = []

        # Since we are in a given SE, the LFN is a unique key
        lfnRepIDs = {}
        for replicaID in seReplicaIDs:
            lfn = replicaIDs[replicaID]["LFN"]
            lfnRepIDs[lfn] = replicaID

        if lfnRepIDs:
            gLogger.info(
                "StageMonitor.__monitorStorageElementStageRequests: Monitoring %s stage requests for %s."
                % (len(lfnRepIDs), storageElement)
            )
        else:
            gLogger.warn(
                "StageMonitor.__monitorStorageElementStageRequests: No requests to monitor for %s." % storageElement
            )
            return
        oAccounting = DataOperation()
        oAccounting.setStartTime()

        res = StorageElement(storageElement, plugins=self.storagePlugins).getFileMetadata(lfnRepIDs)
        if not res["OK"]:
            gLogger.error(
                "StageMonitor.__monitorStorageElementStageRequests: Completely failed to monitor stage requests for replicas",
                res["Message"],
            )
            return
        prestageStatus = res["Value"]

        accountingDict = self.__newAccountingDict(storageElement)

        for lfn, reason in prestageStatus["Failed"].items():
            accountingDict["TransferTotal"] += 1
            if re.search("File does not exist", reason):
                gLogger.error(
                    "StageMonitor.__monitorStorageElementStageRequests: LFN did not exist in the StorageElement", lfn
                )
                terminalReplicaIDs[lfnRepIDs[lfn]] = "LFN did not exist in the StorageElement"
        for lfn, metadata in prestageStatus["Successful"].items():
            if not metadata:
                continue
            staged = metadata.get("Cached", metadata["Accessible"])
            if staged:
                accountingDict["TransferTotal"] += 1
                accountingDict["TransferOK"] += 1
                accountingDict["TransferSize"] += metadata["Size"]
                stagedReplicas.append(lfnRepIDs[lfn])
            elif staged is not None:
                oldRequests.append(lfnRepIDs[lfn])  # only ReplicaIDs

        oAccounting.setValuesFromDict(accountingDict)
        oAccounting.setEndTime()
        gDataStoreClient.addRegister(oAccounting)

        # Update the states of the replicas in the database
        if terminalReplicaIDs:
            gLogger.info(
                "StageMonitor.__monitorStorageElementStageRequests: %s replicas are terminally failed."
                % len(terminalReplicaIDs)
            )
            res = self.stagerClient.updateReplicaFailure(terminalReplicaIDs)
            if not res["OK"]:
                gLogger.error(
                    "StageMonitor.__monitorStorageElementStageRequests: Failed to update replica failures.",
                    res["Message"],
                )
        if stagedReplicas:
            gLogger.info(
                "StageMonitor.__monitorStorageElementStageRequests: %s staged replicas to be updated."
                % len(stagedReplicas)
            )
            res = self.stagerClient.setStageComplete(stagedReplicas)
            if not res["OK"]:
                gLogger.error(
                    "StageMonitor.__monitorStorageElementStageRequests: Failed to updated staged replicas.",
                    res["Message"],
                )
            res = self.stagerClient.updateReplicaStatus(stagedReplicas, "Staged")
            if not res["OK"]:
                gLogger.error(
                    "StageMonitor.__monitorStorageElementStageRequests: Failed to insert replica status.",
                    res["Message"],
                )
        if oldRequests:
            gLogger.info(
                "StageMonitor.__monitorStorageElementStageRequests: %s old requests will be retried." % len(oldRequests)
            )
            res = self.__wakeupOldRequests(oldRequests)
            if not res["OK"]:
                gLogger.error(
                    "StageMonitor.__monitorStorageElementStageRequests: Failed to wakeup old requests.", res["Message"]
                )
        return

    def __newAccountingDict(self, storageElement):
        """Generate a new accounting Dict"""

        accountingDict = {}
        accountingDict["OperationType"] = "Stage"
        accountingDict["User"] = self.proxyInfoDict["username"]
        accountingDict["Protocol"] = "Stager"
        accountingDict["RegistrationTime"] = 0.0
        accountingDict["RegistrationOK"] = 0
        accountingDict["RegistrationTotal"] = 0
        accountingDict["FinalStatus"] = "Successful"
        accountingDict["Source"] = storageElement
        accountingDict["Destination"] = storageElement
        accountingDict["ExecutionSite"] = siteName()
        accountingDict["TransferTotal"] = 0
        accountingDict["TransferOK"] = 0
        accountingDict["TransferSize"] = 0
        accountingDict["TransferTime"] = self.am_getPollingTime()

        return accountingDict

    def __getStageSubmittedReplicas(self):
        """This obtains the StageSubmitted replicas from the Replicas table and the RequestID
        from the StageRequests table
        """
        res = self.stagerClient.getCacheReplicas({"Status": "StageSubmitted"})
        if not res["OK"]:
            gLogger.error(
                "StageMonitor.__getStageSubmittedReplicas: Failed to get replicas with StageSubmitted status.",
                res["Message"],
            )
            return res
        if not res["Value"]:
            gLogger.debug("StageMonitor.__getStageSubmittedReplicas: No StageSubmitted replicas found to process.")
            return S_OK()
        else:
            gLogger.debug(
                "StageMonitor.__getStageSubmittedReplicas: Obtained %s StageSubmitted replicas(s) to process."
                % len(res["Value"])
            )

        seReplicas = {}
        replicaIDs = res["Value"]
        for replicaID, info in replicaIDs.items():
            storageElement = info["SE"]
            seReplicas.setdefault(storageElement, []).append(replicaID)

        # RequestID was missing from replicaIDs dictionary BUGGY?
        res = self.stagerClient.getStageRequests({"ReplicaID": list(replicaIDs)})
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR("Could not obtain request IDs for replicas %s from StageRequests table" % list(replicaIDs))

        for replicaID, info in res["Value"].items():
            replicaIDs[replicaID]["RequestID"] = info["RequestID"]

        return S_OK({"SEReplicas": seReplicas, "ReplicaIDs": replicaIDs})

    def __wakeupOldRequests(self, oldRequests):
        gLogger.info("StageMonitor.__wakeupOldRequests: Attempting...")
        retryInterval = self.am_getOption("RetryIntervalHour", 2)
        res = self.stagerClient.wakeupOldRequests(oldRequests, retryInterval)
        if not res["OK"]:
            gLogger.error("StageMonitor.__wakeupOldRequests: Failed to resubmit old requests.", res["Message"])
            return res
        return S_OK()



""" Basic unit tests for AuthManager
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

from diraccfg import CFG
from DIRAC import gConfig
from DIRAC.Core.DISET.AuthManager import AuthManager

__RCSID__ = "$Id$"

testSystemsCFG = """
Systems
{
  Service
  {
    Authorization
    {
      Method = NormalUser
      MethodAll = Any
      MethodAuth = Authenticated
      MethodGroup = NormalUser,group:group_test
      MethodGroupOther = NormalUser,group:group_test_other
      MethodAllGroup = Any,group:group_test
      MethodAllGroupOther = Any,group:group_test_other
      MethodAuthGroup = Authenticated,group:group_test
      MethodAuthGroupOther = Authenticated,group:group_test_other
      MethodVO = NormalUser,vo:testVO
      MethodAllVO = Any,vo:testVO
      MethodAuthVO = Authenticated,vo:testVO
      MethodHost = group:hosts
      MethodTrustedHost = TrustedHost,group:hosts
    }
  }
}
"""

testRegistryCFG = """
Registry
{
  VO
  {
    testVO
    {
      VOAdmin = userA
    }
    testVOBad
    {
      VOAdmin = userB
    }
    testVOOther
    {
      VOAdmin = userA
    }
  }
  Users
  {
    userA
    {
      DN = /User/test/DN/CN=userA
    }
    userB
    {
      DN = /User/test/DN/CN=userB
    }
    userS
    {
      DN = /User/test/DN/CN=userS
      Suspended = testVO
    }
  }
  Hosts
  {
    test.hostA.ch
    {
      DN = /User/test/DN/CN=test.hostA.ch
      Properties = TrustedHost
    }
    test.hostB.ch
    {
      DN = /User/test/DN/CN=test.hostB.ch
      Properties = NoTrustedHost
    }
  }
  Groups
  {
    group_test
    {
      Users = userA, userS
      VO = testVO
      Properties = NormalUser
    }
    group_test_other
    {
      Users = userS
      VO = testVOOther
      Properties = NormalUser
    }
    group_bad
    {
      Users = userB
      VO = testVOBad
      Properties = NoProperties
    }
  }
}
"""


class AuthManagerTest(unittest.TestCase):
    """Base class for the Modules test cases"""

    def setUp(self):
        self.authMgr = AuthManager("/Systems/Service/Authorization")
        cfg = CFG()
        cfg.loadFromBuffer(testSystemsCFG)
        gConfig.loadCFG(cfg)
        cfg.loadFromBuffer(testRegistryCFG)
        gConfig.loadCFG(cfg)

        self.noAuthCredDict = {"group": "group_test"}

        self.userCredDict = {"DN": "/User/test/DN/CN=userA", "group": "group_test"}
        self.suspendedOtherVOUserCredDict = {"DN": "/User/test/DN/CN=userS", "group": "group_test_other"}
        self.badUserCredDict = {"DN": "/User/test/DN/CN=userB", "group": "group_bad"}
        self.suspendedUserCredDict = {"DN": "/User/test/DN/CN=userS", "group": "group_test"}
        self.hostCredDict = {"DN": "/User/test/DN/CN=test.hostA.ch", "group": "hosts"}
        self.badHostCredDict = {"DN": "/User/test/DN/CN=test.hostB.ch", "group": "hosts"}

    def tearDown(self):
        pass

    def test_userProperties(self):

        # MethodAll accepts everybody
        result = self.authMgr.authQuery("MethodAll", self.userCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAll", self.noAuthCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAll", self.badUserCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAll", self.suspendedUserCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAll", self.suspendedOtherVOUserCredDict)
        self.assertTrue(result)

        # MethodAuth requires DN to be identified
        result = self.authMgr.authQuery("MethodAuth", self.userCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAuth", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuth", self.badUserCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAuth", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuth", self.suspendedOtherVOUserCredDict)
        self.assertTrue(result)

        # Method requires NormalUser property
        result = self.authMgr.authQuery("Method", self.userCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("Method", self.badUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("Method", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("Method", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("Method", self.suspendedOtherVOUserCredDict)
        self.assertTrue(result)

    def test_userGroup(self):

        # MethodAllGroup accepts everybody from the right group
        result = self.authMgr.authQuery("MethodAllGroup", self.userCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAllGroup", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAllGroup", self.badUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAllGroup", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAllGroup", self.suspendedOtherVOUserCredDict)
        self.assertFalse(result)

        # MethodAllGroupOther accepts everybody from the right group
        result = self.authMgr.authQuery("MethodAllGroupOther", self.userCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAllGroupOther", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAllGroupOther", self.badUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAllGroupOther", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAllGroupOther", self.suspendedOtherVOUserCredDict)
        self.assertTrue(result)

        # MethodAuthGroup requires DN to be identified from the right group
        result = self.authMgr.authQuery("MethodAuthGroup", self.userCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAuthGroup", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuthGroup", self.badUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuthGroup", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuthGroup", self.suspendedOtherVOUserCredDict)
        self.assertFalse(result)

        # MethodAuthGroupOther requires DN to be identified from the right group
        result = self.authMgr.authQuery("MethodAuthGroupOther", self.userCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuthGroupOther", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuthGroupOther", self.badUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuthGroupOther", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuthGroupOther", self.suspendedOtherVOUserCredDict)
        self.assertTrue(result)

        # MethodGroup requires NormalUser property and the right group
        result = self.authMgr.authQuery("MethodGroup", self.userCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodGroup", self.badUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodGroup", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodGroup", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodGroup", self.suspendedOtherVOUserCredDict)
        self.assertFalse(result)

        # MethodGroupOther requires NormalUser property and the right group
        result = self.authMgr.authQuery("MethodGroupOther", self.userCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodGroupOther", self.badUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodGroupOther", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodGroupOther", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodGroupOther", self.suspendedOtherVOUserCredDict)
        self.assertTrue(result)

    def test_userVO(self):

        # MethodAllGroup accepts everybody from the right group
        result = self.authMgr.authQuery("MethodAllVO", self.userCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAllVO", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAllVO", self.badUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAllVO", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAllVO", self.suspendedOtherVOUserCredDict)
        self.assertFalse(result)

        # MethodAuthGroup requires DN to be identified from the right group
        result = self.authMgr.authQuery("MethodAuthVO", self.userCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAuthVO", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuthVO", self.badUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuthVO", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodAuthVO", self.suspendedOtherVOUserCredDict)
        self.assertFalse(result)

        # Method requires NormalUser property and the right group
        result = self.authMgr.authQuery("MethodVO", self.userCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodVO", self.badUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodVO", self.noAuthCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodVO", self.suspendedUserCredDict)
        self.assertFalse(result)
        result = self.authMgr.authQuery("MethodVO", self.suspendedOtherVOUserCredDict)
        self.assertFalse(result)

    def test_hostProperties(self):

        # MethodAll accepts everybody
        result = self.authMgr.authQuery("MethodAll", self.hostCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAll", self.badHostCredDict)
        self.assertTrue(result)

        # MethodAuth requires DN to be identified
        result = self.authMgr.authQuery("MethodAuth", self.hostCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodAuth", self.badHostCredDict)
        self.assertTrue(result)

        # Method requires NormalUser property
        result = self.authMgr.authQuery("Method", self.hostCredDict)
        self.assertFalse(result)

        # MethodHost requires hosts group
        result = self.authMgr.authQuery("MethodHost", self.hostCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodHost", self.badHostCredDict)
        self.assertTrue(result)

        # MethodTrustedHost requires hosts group and TrustedHost property
        result = self.authMgr.authQuery("MethodTrustedHost", self.hostCredDict)
        self.assertTrue(result)
        result = self.authMgr.authQuery("MethodTrustedHost", self.badHostCredDict)
        self.assertFalse(result)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(AuthManagerTest)
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



""" Decorators for DIRAC.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import inspect
import functools
import traceback

__RCSID__ = "$Id$"


def deprecated(reason, onlyOnce=False):
    """A decorator to mark a class or function as deprecated.

    This will cause a warnings to be generated in the usual log if the item
    is used (instantiated or called).

    If the environment variable ``DIRAC_DEPRECATED_FAIL`` is set to a non-empty value, an exception will be
    raised when the function or class is used.

    The decorator can be used before as class or function, giving a reason,
    for example::

      @deprecated("Use functionTwo instead")
      def functionOne(...):

    If `onlyOnce` is set to true then the warning will only be generated on the
    first call or creation of the item. This is useful for things that are
    likely to get called repeatedly (to prevent generating massive log files);
    for example::

      @deprecated("Use otherClass instead", onlyOnce=True)
      class MyOldClass:

    If used on a classmethod, it should be used after the `@classmethod` decorator
    for example::

      @classmethod
      @deprecated("Do not put me before @classmethod")
      def methodX(cls):

    Parameters
    ----------
    reason : str
      Message to display to the user when the deprecated item is used. This should specify
      what should be used instead.
    onlyOnce : bool
      If set, the deprecation warning will only be displayed on the first use.

    Returns
    -------
    function
      A double-function wrapper around the decorated object as required by the python
      interpreter.
    """

    def decFunc(func, clsName=None):
        """Inner function generator.
          Returns a function which wraps the given "func" function,
          which prints a deprecation notice as it is called.
          clsName is used internally for class handling, and should be left
          set to None otherwise.

        Parameters
        ----------
        func : function
          The function to call from the wrapper.
        clsName : string
          If set, the wrapped object is assumed to be the __init__ function of
          a class called "clsName". Set to None for wrapping a normal function.

        Returns
        -------
        function
          A function wrapper which which prints the deprecated warning and calls
          func.
        """

        decFunc.warningEn = True

        if func.__doc__ is None:
            func.__doc__ = "\n\n**Deprecated**: " + reason
        else:
            func.__doc__ += "\n\n**Deprecated**: " + reason

        @functools.wraps(func)
        def innerFunc(*args, **kwargs):
            """Prints a suitable deprectaion notice and calls
            the constructor/function/method.
            All arguments are passed through to the target function.
            """
            # fail calling the function if environment variable is set
            if os.environ.get("DIRAC_DEPRECATED_FAIL", None):
                raise NotImplementedError("ERROR: using deprecated function or class: %s" % reason)
            # Get the details of the deprecated object
            if clsName:
                objName = clsName
                objType = "class"
            else:
                objName = func.__name__
                objType = "object"
                if inspect.isfunction(func):
                    objType = "function"
            if decFunc.warningEn:
                # We take the second to last stack frame,
                # which will be the place which called the deprecated item
                # callDetails is a tuple of (file, lineNum, function, text)
                callDetails = traceback.extract_stack()[-2]
                print("NOTE: %s %s is deprecated (%s)." % (objName, objType, reason))
                print("NOTE:   Used at %s:%u" % (callDetails[0], callDetails[1]))
            if onlyOnce:
                decFunc.warningEn = False
            return func(*args, **kwargs)

        # Classes are special, we can decorate them directly,
        # but then calling super( class, inst ) doesn't work as the reference
        # to class becomes a function. Instead we decorate the class __init__
        # function, but then have to override the name otherwise just "__init__ is
        # deprecated" will be printed.
        if inspect.isclass(func):
            func.__init__ = decFunc(func.__init__, clsName=func.__name__)
            return func
        return innerFunc

    return decFunc


def executeOnlyIf(attrName, returnedError, attrVal=None):
    """A decorator to test the value of the attribute of a class before executing a method.

    We often have classes in DIRAC that sets an attribute to True when they the object
    has been successfuly instanciated. And then each and every method test this parameter
    at the start of it.
    This is yet another (very) poor man solution to avoid using exceptions.

    This decorator will do the test for you.

    Pitty, but it breaks the doc, as functools.wrap does not propagate the default attribute discovery

    :param attrName: the name of the attribute to test. If undefined, equivalent to its value being None
    :param returnedError: what to return if the attribute value is not what we expect
    :param attrVal: if set to anything else than None, we check that the attribute value is what is give.
                    If set to None, we just check evaluate __bool__ on the attribute.

    For example::


      class ExceptionsAreEvil(object):

        _hasBeenInitialized = False

        def __init__(self):
          self._hasBeenInitialized = 'OfCourse'

        @executeOnlyIf("_hasBeenInitialized", S_ERROR("How could I not be initialized ?"))
        def functionOne(...):
          doTheActualWork()

        def stupidMethod(...):

          print "I don't like decorator"
          if not self._hasBeenInitialized:
            return S_ERROR("How could I not be initialized ?")

          finallyDoSomething()


    """

    def specificOnlyIf(meth):
        """onlyIf method applied to a specific case"""

        # This utilities allow to preserve the original help
        # Of the method being decorated
        @functools.wraps(meth)
        def innerFunc(*args, **kwargs):
            """Test the attribute before executing the method"""
            self = args[0]

            # First condition is if we have been given a specific attrVal
            # second condition is if we have not been given it
            if (attrVal is not None and getattr(self, attrName, None) != attrVal) or (
                attrVal is None and not getattr(self, attrName, None)
            ):
                return returnedError

            return meth(*args, **kwargs)

        return innerFunc

    return specificOnlyIf



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# """
# Test_RSS_Command_GOCDBSyncCommand
# """

__RCSID__ = "$Id$"

from datetime import datetime, timedelta

import mock

from DIRAC import gLogger, S_OK
from DIRAC.ResourceStatusSystem.Command.GOCDBSyncCommand import GOCDBSyncCommand

mock_GOCDBClient = mock.MagicMock()
mock_RMClient = mock.MagicMock()
mock_RMClient.addOrModifyDowntimeCache.return_value = S_OK()


"""
Setup
"""
gLogger.setLevel("DEBUG")


def test_instantiate():
    """tests that we can instantiate one object of the tested class"""

    command = GOCDBSyncCommand()
    assert command.__class__.__name__ == "GOCDBSyncCommand"


def test_init():
    """tests that the init method does what it should do"""

    command = GOCDBSyncCommand()
    assert command.args == {"onlyCache": False}
    assert command.apis == {}

    command = GOCDBSyncCommand(clients={"GOCDBClient": mock_GOCDBClient})
    assert command.args == {"onlyCache": False}
    assert command.apis == {"GOCDBClient": mock_GOCDBClient}


def test_doNew():
    """tests the doNew method"""

    now = datetime.utcnow()
    resFromDB = {
        "OK": True,
        "Value": (
            (
                now - timedelta(hours=2),
                "dummy.host1.dummy",
                "https://a1.domain",
                now + timedelta(hours=3),
                "dummy.host.dummy",
                now - timedelta(hours=2),
                "maintenance",
                "OUTAGE",
                now,
                "Resource",
                "APEL",
            ),
            (
                now - timedelta(hours=2),
                "dummy.host2.dummy",
                "https://a2.domain",
                now + timedelta(hours=3),
                "dummy.host2.dummy",
                now - timedelta(hours=2),
                "maintenance",
                "OUTAGE",
                now,
                "Resource",
                "CREAM",
            ),
        ),
        "Columns": [
            "StartDate",
            "DowntimeID",
            "Link",
            "EndDate",
            "Name",
            "DateEffective",
            "Description",
            "Severity",
            "LastCheckTime",
            "Element",
            "GOCDBServiceType",
        ],
    }

    resFromGOCDBclient = {
        "OK": True,
        "Value": """<?xml version="1.0" encoding="UTF-8"?>
                    <results>
                      <DOWNTIME ID="dummy.host1.dummy" PRIMARY_KEY="dummy.host1.dummy" CLASSIFICATION="SCHEDULED">
                        <PRIMARY_KEY>dummy.host1.dummy</PRIMARY_KEY>
                        <HOSTNAME>dummy.host1.dummy</HOSTNAME>
                        <SERVICE_TYPE>gLExec</SERVICE_TYPE>
                        <ENDPOINT>dummy.host1.dummy</ENDPOINT>
                        <HOSTED_BY>dummy.host1.dummy</HOSTED_BY>
                        <GOCDB_PORTAL_URL>https://a1.domain</GOCDB_PORTAL_URL>
                        <AFFECTED_ENDPOINTS/>
                        <SEVERITY>OUTAGE</SEVERITY>
                        <DESCRIPTION>Network connectivity problems</DESCRIPTION>
                        <INSERT_DATE>1473460659</INSERT_DATE>
                        <START_DATE>1473547200</START_DATE>
                        <END_DATE>1473677747</END_DATE>
                        <FORMATED_START_DATE>2016-09-10 22:40</FORMATED_START_DATE>
                        <FORMATED_END_DATE>2016-09-12 10:55</FORMATED_END_DATE>
                      </DOWNTIME>
                    </results>""",
    }

    mock_RMClient.selectDowntimeCache.return_value = resFromDB
    mock_GOCDBClient.getHostnameDowntime.return_value = resFromGOCDBclient
    command = GOCDBSyncCommand({"ResourceManagementClient": mock_RMClient, "GOCDBClient": mock_GOCDBClient})

    res = command.doNew()
    assert res["OK"] is False



""" Accounting class to stores network metrics gathered by perfSONARs.

    Filled by "Accounting/Network" agent
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.AccountingSystem.Client.Types.BaseAccountingType import BaseAccountingType


class Network(BaseAccountingType):
    """
    Accounting type to stores network metrics gathered by perfSONARs.
    """

    def __init__(self):
        super(Network, self).__init__()

        # IPv6 address has up to 45 chars
        self.definitionKeyFields = [
            ("SourceIP", "VARCHAR(50)"),
            ("DestinationIP", "VARCHAR(50)"),
            ("SourceHostName", "VARCHAR(50)"),
            ("DestinationHostName", "VARCHAR(50)"),
            ("Source", "VARCHAR(50)"),
            ("Destination", "VARCHAR(50)"),
        ]

        self.definitionAccountingFields = [
            ("Jitter", "FLOAT"),
            ("OneWayDelay", "FLOAT"),
            ("PacketLossRate", "TINYINT UNSIGNED"),
        ]
        self.bucketsLength = [
            (86400 * 3, 900),  # <3d = 15m
            (86400 * 8, 3600),  # <1w+1d = 1h
            (15552000, 86400),  # >1w+1d <6m = 1d
            (31104000, 604800),  # >6m = 1w
        ]
        self.checkType()



""" This is for transfers what RPCClient is for RPC calls
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import File
from DIRAC.Core.DISET.private.BaseClient import BaseClient
from DIRAC.Core.DISET.private.FileHelper import FileHelper


class TransferClient(BaseClient):
    def _sendTransferHeader(self, actionName, fileInfo):
        """
        Send the header of the transfer

        :type actionName: string
        :param actionName: Action to execute
        :type fileInfo: tuple
        :param fileInfo: Information of the target file/bulk
        :return: S_OK/S_ERROR
        """
        retVal = self._connect()
        if not retVal["OK"]:
            return retVal
        trid, transport = retVal["Value"]
        try:
            # FFC -> File from Client
            retVal = self._proposeAction(transport, ("FileTransfer", actionName))
            if not retVal["OK"]:
                return retVal
            # We need to convert to list
            retVal = transport.sendData(S_OK(list(fileInfo)))
            if not retVal["OK"]:
                return retVal
            retVal = transport.receiveData()
            if not retVal["OK"]:
                return retVal
            return S_OK((trid, transport))
        except Exception as e:
            self._disconnect(trid)
            return S_ERROR("Cound not request transfer: %s" % str(e))

    def sendFile(self, filename, fileId, token=""):
        """
        Send a file to server

        :type filename: string / file descriptor / file object
        :param filename: File to send to server
        :type fileId: any
        :param fileId: Identification of the file being sent
        :type token: string
        :param token: Optional token for the file
        :return: S_OK/S_ERROR
        """
        fileHelper = FileHelper()
        if "NoCheckSum" in token:
            fileHelper.disableCheckSum()
        retVal = fileHelper.getFileDescriptor(filename, "r")
        if not retVal["OK"]:
            return retVal
        fd = retVal["Value"]
        retVal = self._sendTransferHeader("FromClient", (fileId, token, File.getSize(filename)))
        if not retVal["OK"]:
            return retVal
        trid, transport = retVal["Value"]
        try:
            fileHelper.setTransport(transport)
            retVal = fileHelper.FDToNetwork(fd)
            if not retVal["OK"]:
                return retVal
            retVal = transport.receiveData()
            return retVal
        finally:
            self._disconnect(trid)

    def receiveFile(self, filename, fileId, token=""):
        """
        Receive a file from the server

        :type filename: string / file descriptor / file object
        :param filename: File to receive from server
        :type fileId: any
        :param fileId: Identification of the file being received
        :type token: string
        :param token: Optional token for the file
        :return: S_OK/S_ERROR
        """
        fileHelper = FileHelper()
        if "NoCheckSum" in token:
            fileHelper.disableCheckSum()
        retVal = fileHelper.getDataSink(filename)
        if not retVal["OK"]:
            return retVal
        dS = retVal["Value"]
        closeAfterUse = retVal["closeAfterUse"]
        retVal = self._sendTransferHeader("ToClient", (fileId, token))
        if not retVal["OK"]:
            return retVal
        trid, transport = retVal["Value"]
        try:
            fileHelper.setTransport(transport)
            retVal = fileHelper.networkToDataSink(dS)
            if not retVal["OK"]:
                return retVal
            retVal = transport.receiveData()
            if closeAfterUse:
                dS.close()
            return retVal
        finally:
            self._disconnect(trid)

    def __checkFileList(self, fileList):
        bogusEntries = []
        for entry in fileList:
            if not os.path.exists(entry):
                bogusEntries.append(entry)
        return bogusEntries

    def sendBulk(self, fileList, bulkId, token="", compress=True, bulkSize=-1, onthefly=True):
        """
        Send a bulk of files to server

        :type fileList: list of ( string / file descriptor / file object )
        :param fileList: Files to send to server
        :type bulkId: any
        :param bulkId: Identification of the files being sent
        :type token: string
        :param token: Token for the bulk
        :type compress: boolean
        :param compress: Enable compression for the bulk. By default its True
        :type bulkSize: integer
        :param bulkSize: Optional size of the bulk
        :return: S_OK/S_ERROR
        """
        bogusEntries = self.__checkFileList(fileList)
        if bogusEntries:
            return S_ERROR("Some files or directories don't exist :\n\t%s" % "\n\t".join(bogusEntries))
        if compress:
            bulkId = "%s.tar.bz2" % bulkId
        else:
            bulkId = "%s.tar" % bulkId
        retVal = self._sendTransferHeader("BulkFromClient", (bulkId, token, bulkSize))
        if not retVal["OK"]:
            return retVal
        trid, transport = retVal["Value"]
        try:
            fileHelper = FileHelper(transport)
            retVal = fileHelper.bulkToNetwork(fileList, compress, onthefly)
            if not retVal["OK"]:
                return retVal
            retVal = transport.receiveData()
            return retVal
        finally:
            self._disconnect(trid)

    def receiveBulk(self, destDir, bulkId, token="", compress=True):
        """
        Receive a bulk of files from server

        :type destDir: list of ( string / file descriptor / file object )
        :param destDir: Files to receive from server
        :type bulkId: any
        :param bulkId: Identification of the files being received
        :type token: string
        :param token: Token for the bulk
        :type compress: boolean
        :param compress: Enable compression for the bulk. By default its True
        :return: S_OK/S_ERROR
        """
        if not os.path.isdir(destDir):
            return S_ERROR("%s is not a directory for bulk receival" % destDir)
        if compress:
            bulkId = "%s.tar.bz2" % bulkId
        else:
            bulkId = "%s.tar" % bulkId
        retVal = self._sendTransferHeader("BulkToClient", (bulkId, token))
        if not retVal["OK"]:
            return retVal
        trid, transport = retVal["Value"]
        try:
            fileHelper = FileHelper(transport)
            retVal = fileHelper.networkToBulk(destDir, compress)
            if not retVal["OK"]:
                return retVal
            retVal = transport.receiveData()
            return retVal
        finally:
            self._disconnect(trid)

    def listBulk(self, bulkId, token="", compress=True):
        """
        List the contents of a bulk

        :type bulkId: any
        :param bulkId: Identification of the bulk to list
        :type token: string
        :param token: Token for the bulk
        :type compress: boolean
        :param compress: Enable compression for the bulk. By default its True
        :return: S_OK/S_ERROR
        """
        if compress:
            bulkId = "%s.tar.bz2" % bulkId
        else:
            bulkId = "%s.tar" % bulkId
        trid = None
        retVal = self._sendTransferHeader("ListBulk", (bulkId, token))
        if not retVal["OK"]:
            return retVal
        trid, transport = retVal["Value"]
        try:
            response = transport.receiveData(1048576)
            return response
        finally:
            self._disconnect(trid)



#!/usr/bin/env python
########################################################################
# File :   dirac-admin-voms-sync
# Author : Andrei Tsaregorodtsev
########################################################################
"""
Synchronize VOMS user data with the DIRAC Registry
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from DIRAC import gLogger, exit as DIRACExit, S_OK
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.ConfigurationSystem.Client.VOMS2CSSynchronizer import VOMS2CSSynchronizer
from DIRAC.Core.Utilities.Proxy import executeWithUserProxy
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOOption


dryRun = False
voName = None


def setDryRun(value):
    global dryRun
    dryRun = True
    return S_OK()


def setVO(value):
    global voName
    voName = value
    return S_OK()


@Script()
def main():
    Script.registerSwitch("V:", "vo=", "VO name", setVO)
    Script.registerSwitch("D", "dryRun", "Dry run", setDryRun)
    Script.parseCommandLine(ignoreErrors=True)

    @executeWithUserProxy
    def syncCSWithVOMS(vomsSync):
        return vomsSync.syncCSWithVOMS()

    voAdminUser = getVOOption(voName, "VOAdmin")
    voAdminGroup = getVOOption(voName, "VOAdminGroup", getVOOption(voName, "DefaultGroup"))

    vomsSync = VOMS2CSSynchronizer(voName)
    result = syncCSWithVOMS(  # pylint: disable=unexpected-keyword-arg
        vomsSync, proxyUserName=voAdminUser, proxyUserGroup=voAdminGroup
    )
    if not result["OK"]:
        gLogger.error("Failed to synchronize user data")
        DIRACExit(-1)

    resultDict = result["Value"]
    newUsers = resultDict.get("NewUsers", [])
    modUsers = resultDict.get("ModifiedUsers", [])
    delUsers = resultDict.get("DeletedUsers", [])
    susUsers = resultDict.get("SuspendedUsers", [])
    gLogger.notice(
        "\nUser results: new %d, modified %d, deleted %d, new/suspended %d"
        % (len(newUsers), len(modUsers), len(delUsers), len(susUsers))
    )

    for msg in resultDict["AdminMessages"]["Info"]:
        gLogger.notice(msg)

    csapi = resultDict.get("CSAPI")
    if csapi and csapi.csModified:
        if dryRun:
            gLogger.notice("There are changes to Registry ready to commit, skipped because of dry run")
        else:
            yn = six.moves.input("There are changes to Registry ready to commit, do you want to proceed ? [Y|n]:")
            if yn == "" or yn[0].lower() == "y":
                result = csapi.commitChanges()
                if not result["OK"]:
                    gLogger.error("Could not commit configuration changes", result["Message"])
                else:
                    gLogger.notice("Registry changes committed for VO %s" % voName)
            else:
                gLogger.notice("Registry changes are not committed")
    else:
        gLogger.notice("No changes to Registry for VO %s" % voName)

    result = vomsSync.getVOUserReport()
    if not result["OK"]:
        gLogger.error("Failed to generate user data report")
        DIRACExit(-1)

    gLogger.notice("\n" + result["Value"])


if __name__ == "__main__":
    main()



""" LogPolicyResultAction

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from DIRAC.ResourceStatusSystem.PolicySystem.Actions.BaseAction import BaseAction
from DIRAC.ResourceStatusSystem.Utilities import Utils

ResourceManagementClient = getattr(
    Utils.voimport("DIRAC.ResourceStatusSystem.Client.ResourceManagementClient"), "ResourceManagementClient"
)


class LogPolicyResultAction(BaseAction):
    """
    Action that registers on the database a new entry per policy result in the
    list singlePolicyResults.
    """

    def __init__(self, name, decisionParams, enforcementResult, singlePolicyResults, clients=None):

        super(LogPolicyResultAction, self).__init__(
            name, decisionParams, enforcementResult, singlePolicyResults, clients
        )

        if clients is not None and "ResourceManagementClient" in clients:
            self.rmClient = clients["ResourceManagementClient"]
        else:
            self.rmClient = ResourceManagementClient()

    def run(self):
        """
        Checks it has the parameters it needs and tries to addOrModify in the
        database.
        """

        element = self.decisionParams["element"]
        if element is None:
            return S_ERROR("element should not be None")

        name = self.decisionParams["name"]
        if name is None:
            return S_ERROR("name should not be None")

        statusType = self.decisionParams["statusType"]
        if statusType is None:
            return S_ERROR("statusType should not be None")

        for singlePolicyResult in self.singlePolicyResults:

            status = singlePolicyResult["Status"]
            if status is None:
                return S_ERROR("status should not be None")

            reason = singlePolicyResult["Reason"]
            if reason is None:
                return S_ERROR("reason should not be None")

            policyName = singlePolicyResult["Policy"]["name"]
            if policyName is None:
                return S_ERROR("policyName should not be None")

            vo = singlePolicyResult.get("VO")
            # Truncate reason to fit in database column
            reason = (reason[:508] + "..") if len(reason) > 508 else reason

            polUpdateRes = self.rmClient.addOrModifyPolicyResult(
                element=element,
                name=name,
                vO=vo,
                policyName=policyName,
                statusType=statusType,
                status=status,
                reason=reason,
            )

            if not polUpdateRes["OK"]:
                return polUpdateRes

        return S_OK()


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import re
import os
import DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.Core.Utilities import List
from DIRAC.Core.Utilities.Extensions import extensionsByPriority
from DIRAC.Core.Utilities.ObjectLoader import loadObjects


class MessageFactory(object):
    def __init__(self):
        self.__definitions = {}
        self.__svcHandlers = {}

    def createMessage(self, serviceName, msgName, attrs=None):
        result = self.__loadMessagesFromService(serviceName)
        if not result["OK"]:
            return result
        if msgName not in self.__definitions[serviceName]:
            return S_ERROR("Could not find message definition %s for service %s" % (msgName, serviceName))
        msgObj = Message(msgName, self.__definitions[serviceName][msgName])
        if attrs is not None:
            result = msgObj.loadAttrs(attrs)
            if not result["OK"]:
                return result
        return S_OK(msgObj)

    def getMessagesForService(self, serviceName):
        result = self.__loadMessagesFromService(serviceName)
        if not result["OK"]:
            return result
        return S_OK(tuple(self.__definitions[serviceName]))

    def messageExists(self, serviceName, msgName):
        result = self.getMessagesForService(serviceName)
        if not result["OK"]:
            return False
        return msgName in result["Value"]

    def __loadHandler(self, serviceName):
        # TODO: Load handlers as the Service does (1. CS 2. SysNameSystem/Service/servNameHandler.py)
        sL = List.fromChar(serviceName, "/")
        if len(sL) != 2:
            return S_ERROR("Service name is not valid: %s" % serviceName)
        sysName = sL[0]
        svcHandlerName = "%sHandler" % sL[1]
        loadedObjs = loadObjects("%sSystem/Service" % sysName, reFilter=re.compile(r"^%s\.py$" % svcHandlerName))
        if svcHandlerName not in loadedObjs:
            return S_ERROR("Could not find %s for getting messages definition" % serviceName)
        return S_OK(loadedObjs[svcHandlerName])

    def __loadMessagesFromService(self, serviceName):
        if serviceName in self.__definitions:
            return S_OK()
        if serviceName not in self.__svcHandlers:
            result = self.__loadHandler(serviceName)
            if not result["OK"]:
                return result
            self.__svcHandlers[serviceName] = result["Value"]
        self.__definitions[serviceName] = {}
        handlerClass = self.__svcHandlers[serviceName]
        # Load message definition starting with ancestors, children override the ancestors as usual
        result = self.__loadMessagesForAncestry(handlerClass)
        if not result["OK"]:
            return result
        msgDefs = result["Value"]
        if not msgDefs:
            return S_ERROR("%s does not have messages defined" % serviceName)
        self.__definitions[serviceName] = msgDefs
        return S_OK()

    def __loadMessagesForAncestry(self, handlerClass):
        finalDefs = {}
        for ancestor in handlerClass.__bases__:
            result = self.__loadMessagesForAncestry(ancestor)
            if not result["OK"]:
                return result
            ancestorDefs = result["Value"]
            for msgName in ancestorDefs:
                finalDefs[msgName] = ancestorDefs[msgName]
        if "MSG_DEFINITIONS" not in dir(handlerClass):
            return S_OK(finalDefs)
        msgDefs = getattr(handlerClass, "MSG_DEFINITIONS")
        if not isinstance(msgDefs, dict):
            return S_ERROR("Message definitions for service %s is not a dict" % handlerClass.__name__)
        for msgName in msgDefs:
            msgDefDict = msgDefs[msgName]
            if not isinstance(msgDefDict, dict):
                return S_ERROR("Type of message definition has to be a dict")
            finalDefs[msgName] = msgDefDict
        return S_OK(finalDefs)


class Message(object):

    DEFAULTWAITFORACK = False

    def __init__(self, msgName, msgDefDict):
        self.__name = msgName
        self.__fDef = {}
        self.__order = []
        self.__values = {}
        self.__msgClient = None
        self.__waitForAck = Message.DEFAULTWAITFORACK
        for fName in msgDefDict:
            fType = msgDefDict[fName]
            if fType is not None and not isinstance(fType, (list, tuple)):
                self.__fDef[fName] = (fType,)
            else:
                self.__fDef[fName] = fType
            self.__order.append((fName.lower(), fName))
        self.__order = [oF[1] for oF in sorted(self.__order)]
        # "Enable" set attr
        self.__locked = True

    def setMsgClient(self, msgClient):
        self.__msgClient = msgClient

    @property
    def msgClient(self):
        return self.__msgClient

    def isOK(self):
        for k in self.__order:
            if k not in self.__values:
                return False
            if self.__fDef[k] is not None and not isinstance(self.__values[k], self.__fDef[k]):
                return False
        return True

    def setWaitForAck(self, bA):
        self.__waitForAck = bA

    def getWaitForAck(self):
        return self.__waitForAck

    def getName(self):
        return self.__name

    def getAttrNames(self):
        return self.__order

    def loadAttrsFromDict(self, dataDict):
        if not isinstance(dataDict, dict):
            return S_ERROR("Params have to be a dict")
        for k in dataDict:
            try:
                setattr(self, k, dataDict[k])
            except AttributeError as e:
                return S_ERROR(str(e))
        return S_OK()

    def dumpAttrs(self):
        try:
            return S_OK([self.__waitForAck, [self.__values[k] for k in self.__order]])
        except Exception as e:
            return S_ERROR("Could not dump message: %s doesn't have a value" % e)

    def loadAttrs(self, data):
        if not isinstance(data, (list, tuple)) and len(data) != 2:
            return S_ERROR("Data is not valid to be loaded as message")
        self.setWaitForAck(data[0])
        data = data[1]
        if not isinstance(data, (list, tuple)):
            return S_ERROR("Data is not valid to be loaded as message")
        for iP in range(len(self.__order)):
            try:
                v = data[iP]
            except IndexError:
                return S_ERROR("Data is too short!")
            k = self.__order[iP]
            try:
                setattr(self, k, v)
            except AttributeError as e:
                return S_ERROR(str(e))
        return S_OK()

    def __str__(self):
        msgStr = ["<Message %s (" % self.__name]
        for k in self.__order:
            if k in self.__values:
                v = str(self.__values[k])
                if len(v) > 5:
                    v = "%s..." % v[:5]
                msgStr.append("%s->%s" % (k, v))
            else:
                msgStr.append("%s" % k)
        if self.isOK():
            msgStr.append(") OK>")
        else:
            msgStr.append(") >")
        return " ".join(msgStr)

    def __repr__(self):
        return self.__str__()

    def __setattr__(self, k, v):
        # Intialization
        if "_Message__locked" not in self.__dict__:
            object.__setattr__(self, k, v)
            return
        # Normal assignment
        if k in self.__dict__:
            object.__setattr__(self, k, v)
            return
        if k not in self.__order:
            raise AttributeError("%s is not valid for message %s" % (k, self.__name))
        if self.__fDef[k] is not None and not isinstance(v, self.__fDef[k]):
            raise AttributeError(
                "%s is to be of type %s for attr %s, and is of type %s" % (v, self.__fDef[k], k, type(v))
            )
        self.__values[k] = v

    def __getattr__(self, k):
        if k not in self.__values:
            if k not in self.__fDef:
                raise AttributeError("No %s attribute for message %s" % (k, self.__name))
            raise AttributeError("%s has no value" % k)
        return self.__values[k]


#
# MessageContainer
#


class DummyMessage:
    def __init__(self, msg):
        self.msg = msg

    def getName(self):
        return self.msg["name"]

    def dumpAttrs(self):
        return S_OK(self.msg["attrs"])

    def isOK(self):
        return True

    def getWaitForAck(self):
        return self.msg["attrs"][0]



""" Test class for LSFResourceUsage utility
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import pytest

from DIRAC import S_OK, gLogger

from DIRAC.Resources.Computing.BatchSystems.TimeLeft.PBSResourceUsage import PBSResourceUsage

gLogger.setLevel("DEBUG")

RRCKI_OUT = """
Job Id: 55755440.seer.t1.grid.kiae.ru
    Job_Name = igyLDmzixXznYTGZXoIFMvcqIyasxmGoZvymfcHKDmABFKDmVYviIm
    Job_Owner = lhcbpilot0000@calc1.t1.grid.kiae.ru
    resources_used.cput = 43:02:47
    resources_used.mem = 1611016kb
    resources_used.vmem = 3636260kb
    resources_used.walltime = 43:22:30
    job_state = R
    queue = lhcb
    server = seer.t1.grid.kiae.ru
    Checkpoint = u
    ctime = Wed Aug 11 10:20:59 2021
    Error_Path = calc1.t1.grid.kiae.ru:/shared/sandbox/04/igyLDmzixXznYTGZXoIF
        MvcqIyasxmGoZvymfcHKDmABFKDmVYviIm.comment
    exec_host = n175.t1.grid.kiae.ru/17
    Hold_Types = n
    Join_Path = eo
    Keep_Files = n
    Mail_Points = n
    mtime = Wed Aug 11 17:13:13 2021
    Output_Path = calc1.t1.grid.kiae.ru:/shared/sandbox/04/igyLDmzixXznYTGZXoI
        FMvcqIyasxmGoZvymfcHKDmABFKDmVYviIm/igyLDmzixXznYTGZXoIFMvcqIyasxmGoZv
        ymfcHKDmABFKDmVYviIm.o55755440
    Priority = -1
    qtime = Wed Aug 11 10:20:59 2021
    Rerunable = False
    Resource_List.cput = 100:00:00
    Resource_List.nice = 10
    Resource_List.nodect = 1
    Resource_List.nodes = 1
    Resource_List.walltime = 120:00:00
    session_id = 20422
    Shell_Path_List = /bin/bash
    stagein = HOME@calc1.t1.grid.kiae.ru:/shared/sandbox/04/igyLDmzixXznYTG
        ZXoIFMvcqIyasxmGoZvymfcHKDmABFKDmVYviIm
    stageout = HOME@igyLDmzixXznYTGZXoIFMvcqIyasxmGoZvymfcHKDmABFKDmVYviIm@
        calc1.t1.grid.kiae.ru:/shared/sandbox/04
    Variable_List = PBS_O_QUEUE=lhcb,PBS_O_HOST=calc1.t1.grid.kiae.ru,
        PBS_O_HOME=/,PBS_O_LANG=en_US.UTF-8,
        PBS_O_PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin,
        PBS_SERVER=seer.t1.grid.kiae.ru,
        PBS_O_WORKDIR=/shared/sandbox/04/igyLDmzixXznYTGZXoIFMvcqIyasxmGoZvym
        fcHKDmABFKDmVYviIm
    etime = Wed Aug 11 10:20:59 2021
    submit_args = -r n -S /bin/bash -m n
    start_time = Wed Aug 11 17:13:12 2021
    Walltime.Remaining = 275789
    start_count = 1
    fault_tolerant = False
    submit_host = calc1.t1.grid.kiae.ru
    init_work_dir = /shared/sandbox/04/igyLDmzixXznYTGZXoIFMvcqIyasxmGoZvymfcH
        KDmABFKDmVYviIm
    resource_limits.pvmem = 6gb
    resource_limits.vmem = 8gb
    kill_gracetime = 60000
    min_nice = 19
"""


@pytest.mark.parametrize(
    "runCommandResult, \
    cpuLimitExpected, wallClockLimitExpected",
    [
        (
            [S_OK(RRCKI_OUT)],
            154967.0,
            156150.0,
        )
    ],
)
def test_getResourcUsage(
    mocker,
    runCommandResult,
    cpuLimitExpected,
    wallClockLimitExpected,
):
    mocker.patch(
        "DIRAC.Resources.Computing.BatchSystems.TimeLeft.PBSResourceUsage.runCommand",
        side_effect=runCommandResult,
    )
    mocker.patch("os.path.isfile", return_value=True)
    mocker.patch.dict(os.environ, {"PBS_O_QUEUE": "lhcb", "PBS_O_QPATH": "/some/path"})

    pbsRU = PBSResourceUsage()
    pbsRU.jobID = "55755440.seer.t1.grid.kiae.ru"
    res = pbsRU.getResourceUsage()
    assert res["OK"], res["Message"]
    assert len(res["Value"]) == 4
    assert res["Value"]["CPU"] == cpuLimitExpected  # pylint: disable=invalid-sequence-index
    assert res["Value"]["WallClock"] == wallClockLimitExpected  # pylint: disable=invalid-sequence-index



""" VOMSService class encapsulates connection to the VOMS service for a given VO
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import requests

from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Security.Locations import getProxyLocation, getCAsLocation
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOOption
from DIRAC.ConfigurationSystem.Client.Helpers.CSGlobals import getVO


class VOMSService(object):
    def __init__(self, vo=None):
        """c'tor

        :param str vo: name of the virtual organization (community)
        """

        if vo is None:
            vo = getVO()
        if not vo:
            raise Exception("No VO name given")

        self.vo = vo
        self.vomsVO = getVOOption(vo, "VOMSName")
        if not self.vomsVO:
            raise Exception("Can not get VOMS name for VO %s" % vo)

        self.urls = []
        result = gConfig.getSections("/Registry/VO/%s/VOMSServers" % self.vo)
        if result["OK"]:
            for server in result["Value"]:
                gLogger.verbose("Adding 'https://%s:8443/voms/%s/apiv2/users'" % (server, self.vomsVO))
                self.urls.append("https://%s:8443/voms/%s/apiv2/users" % (server, self.vomsVO))
        else:
            gLogger.error("Section '/Registry/VO/%s/VOMSServers' not found" % self.vo)

        self.userDict = None

    def attGetUserNickname(self, dn, _ca=None):
        """Get user nickname for a given DN if any

        :param str dn: user DN
        :param str _ca: CA, kept for backward compatibility
        :return:  S_OK with Value: nickname
        """

        if self.userDict is None:
            result = self.getUsers()
            if not result["OK"]:
                return result

        uDict = self.userDict.get(dn)
        if not uDict:
            return S_ERROR(DErrno.EVOMS, "No nickname defined")
        nickname = uDict.get("nickname")
        if not nickname:
            return S_ERROR(DErrno.EVOMS, "No nickname defined")
        return S_OK(nickname)

    def getUsers(self):
        """Get all the users of the VOMS VO with their detailed information

        :return: user dictionary keyed by the user DN
        """

        if not self.urls:
            return S_ERROR(DErrno.ENOAUTH, "No VOMS server defined")

        userProxy = getProxyLocation()
        caPath = getCAsLocation()
        rawUserList = []
        result = None
        for url in self.urls:
            rawUserList = []
            startIndex = 0
            result = None
            error = None
            urlDone = False
            while not urlDone:
                try:
                    result = requests.get(
                        url,
                        headers={"X-VOMS-CSRF-GUARD": "y"},
                        cert=userProxy,
                        verify=caPath,
                        params={"startIndex": str(startIndex), "pageSize": "100"},
                    )
                except requests.ConnectionError as exc:
                    error = "%s:%s" % (url, repr(exc))
                    urlDone = True
                    continue

                if result.status_code != 200:
                    error = "Failed to contact the VOMS server: %s" % result.text
                    urlDone = True
                    continue

                userList = result.json()["result"]
                rawUserList.extend(userList)
                if len(userList) < 100:
                    urlDone = True
                startIndex += 100

            # This URL did not work, try another one
            if error:
                continue
            else:
                break

        if error:
            return S_ERROR(DErrno.ENOAUTH, "Failed to contact the VOMS server: %s" % error)

        # We have got the user info, reformat it
        resultDict = {}
        for user in rawUserList:
            for cert in user["certificates"]:
                dn = cert["subjectString"]
                resultDict[dn] = user
                resultDict[dn]["CA"] = cert["issuerString"]
                resultDict[dn]["certSuspended"] = cert.get("suspended")
                resultDict[dn]["suspended"] = user.get("suspended")
                resultDict[dn]["mail"] = user.get("emailAddress")
                resultDict[dn]["Roles"] = user.get("fqans")
                attributes = user.get("attributes")
                if attributes:
                    for attribute in user.get("attributes", []):
                        if attribute.get("name") == "nickname":
                            resultDict[dn]["nickname"] = attribute.get("value")

        self.userDict = dict(resultDict)
        return S_OK(resultDict)



#!/usr/bin/env python
"""
Enable using one or more Storage Elements

Example:
  $ dirac-admin-allow-se M3PEC-disk
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    read = False
    write = False
    check = False
    remove = False
    site = ""
    mute = False

    Script.registerSwitch("r", "AllowRead", "     Allow only reading from the storage element")
    Script.registerSwitch("w", "AllowWrite", "     Allow only writing to the storage element")
    Script.registerSwitch("k", "AllowCheck", "     Allow only check access to the storage element")
    Script.registerSwitch("v", "AllowRemove", "    Allow only remove access to the storage element")
    Script.registerSwitch("a", "All", "    Allow all access to the storage element")
    Script.registerSwitch("m", "Mute", "     Do not send email")
    Script.registerSwitch("S:", "Site=", "     Allow all SEs associated to site")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["seGroupList: list of SEs or comma-separated SEs"])

    switches, ses = Script.parseCommandLine(ignoreErrors=True)

    for switch in switches:
        if switch[0].lower() in ("r", "allowread"):
            read = True
        if switch[0].lower() in ("w", "allowwrite"):
            write = True
        if switch[0].lower() in ("k", "allowcheck"):
            check = True
        if switch[0].lower() in ("v", "allowremove"):
            remove = True
        if switch[0].lower() in ("a", "all"):
            read = True
            write = True
            check = True
            remove = True
        if switch[0].lower() in ("m", "mute"):
            mute = True
        if switch[0].lower() in ("s", "site"):
            site = switch[1]

    # imports
    from DIRAC import gConfig, gLogger
    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
    from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites
    from DIRAC.Core.Security.ProxyInfo import getProxyInfo
    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
    from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus

    if not (read or write or check or remove):
        # No switch was specified, means we need all of them
        gLogger.notice("No option given, all accesses will be allowed if they were not")
        read = True
        write = True
        check = True
        remove = True

    ses = resolveSEGroup(ses)
    diracAdmin = DiracAdmin()
    errorList = []
    setup = gConfig.getValue("/DIRAC/Setup", "")
    if not setup:
        print("ERROR: Could not contact Configuration Service")
        DIRAC.exit(2)

    res = getProxyInfo()
    if not res["OK"]:
        gLogger.error("Failed to get proxy information", res["Message"])
        DIRAC.exit(2)

    userName = res["Value"].get("username")
    if not userName:
        gLogger.error("Failed to get username for proxy")
        DIRAC.exit(2)

    if site:
        res = getSites()
        if not res["OK"]:
            gLogger.error(res["Message"])
            DIRAC.exit(-1)
        if site not in res["Value"]:
            gLogger.error("The provided site (%s) is not known." % site)
            DIRAC.exit(-1)
        ses.extend(res["Value"]["SE"].replace(" ", "").split(","))
    if not ses:
        gLogger.error("There were no SEs provided")
        DIRAC.exit()

    STATUS_TYPES = ["ReadAccess", "WriteAccess", "CheckAccess", "RemoveAccess"]
    ALLOWED_STATUSES = ["Unknown", "InActive", "Banned", "Probing", "Degraded"]

    statusAllowedDict = {}
    for statusType in STATUS_TYPES:
        statusAllowedDict[statusType] = []

    statusFlagDict = {}
    statusFlagDict["ReadAccess"] = read
    statusFlagDict["WriteAccess"] = write
    statusFlagDict["CheckAccess"] = check
    statusFlagDict["RemoveAccess"] = remove

    resourceStatus = ResourceStatus()

    res = resourceStatus.getElementStatus(ses, "StorageElement")
    if not res["OK"]:
        gLogger.error("Storage Element %s does not exist" % ses)
        DIRAC.exit(-1)

    reason = "Forced with dirac-admin-allow-se by %s" % userName

    for se, seOptions in res["Value"].items():
        # InActive is used on the CS model, Banned is the equivalent in RSS
        for statusType in STATUS_TYPES:
            if statusFlagDict[statusType]:
                if seOptions.get(statusType) == "Active":
                    gLogger.notice("%s status of %s is already Active" % (statusType, se))
                    continue
                if statusType in seOptions:
                    if not seOptions[statusType] in ALLOWED_STATUSES:
                        gLogger.notice(
                            "%s option for %s is %s, instead of %s"
                            % (statusType, se, seOptions["ReadAccess"], ALLOWED_STATUSES)
                        )
                        gLogger.notice("Try specifying the command switches")
                    else:
                        resR = resourceStatus.setElementStatus(
                            se, "StorageElement", statusType, "Active", reason, userName
                        )
                        if not resR["OK"]:
                            gLogger.fatal(
                                "Failed to update %s %s to Active, exit -" % (se, statusType), resR["Message"]
                            )
                            DIRAC.exit(-1)
                        else:
                            gLogger.notice("Successfully updated %s %s to Active" % (se, statusType))
                            statusAllowedDict[statusType].append(se)

    totalAllowed = 0
    totalAllowedSEs = []
    for statusType in STATUS_TYPES:
        totalAllowed += len(statusAllowedDict[statusType])
        totalAllowedSEs += statusAllowedDict[statusType]
    totalAllowedSEs = list(set(totalAllowedSEs))

    if not totalAllowed:
        gLogger.info("No storage elements were allowed")
        DIRAC.exit(-1)

    if mute:
        gLogger.notice("Email is muted by script switch")
        DIRAC.exit(0)

    subject = "%s storage elements allowed for use" % len(totalAllowedSEs)
    addressPath = "EMail/Production"
    address = Operations().getValue(addressPath, "")

    body = ""
    if read:
        body = "%s\n\nThe following storage elements were allowed for reading:" % body
        for se in statusAllowedDict["ReadAccess"]:
            body = "%s\n%s" % (body, se)
    if write:
        body = "%s\n\nThe following storage elements were allowed for writing:" % body
        for se in statusAllowedDict["WriteAccess"]:
            body = "%s\n%s" % (body, se)
    if check:
        body = "%s\n\nThe following storage elements were allowed for checking:" % body
        for se in statusAllowedDict["CheckAccess"]:
            body = "%s\n%s" % (body, se)
    if remove:
        body = "%s\n\nThe following storage elements were allowed for removing:" % body
        for se in statusAllowedDict["RemoveAccess"]:
            body = "%s\n%s" % (body, se)

    if not address:
        gLogger.notice("'%s' not defined in Operations, can not send Mail\n" % addressPath, body)
        DIRAC.exit(0)

    res = diracAdmin.sendMail(address, subject, body)
    gLogger.notice("Notifying %s" % address)
    if res["OK"]:
        gLogger.notice(res["Value"])
    else:
        gLogger.notice(res["Message"])

    DIRAC.exit(0)


if __name__ == "__main__":
    main()
